"""
Module for interacting with tabulated data in ``pyXMIP``.

Notes
-----

For general usage, :py:class:`SourceTable` operates exactly the same as :py:class:`astropy.table.table.Table` with only a few
key exceptions:

1. :py:class:`SourceTable` implements a :py:class:`schema.SourceTableSchema` on top of the existing astropy structure.
2. :py:class:`SourceTable` utilizes its schema to deduce additional information not inherently included in the table.

For general usage information, particularly regarding interacting with, slicing from, and altering tables, we recommend
reading the associated ``astropy`` documentation `here <https://docs.astropy.org/en/stable/table/index.html>`_.
"""
import re
from typing import Generic, Type, TypeVar

import numpy as np
import pandas as pd
from astropy import units as units
from astropy.coordinates import GenericFrame, SkyCoord
from astropy.coordinates.angles.core import Latitude, Longitude
from astropy.io import fits
from astropy.table import Column, Table, TableAttribute

from pyXMIP.schema import SourceTableSchema
from pyXMIP.utilities.core import enforce_units

Instance = TypeVar("Instance")
Value = TypeVar("Value")
Attribute = TypeVar("Attribute")


class SchemaColumn(Generic[Instance, Attribute, Value]):
    """
    Special column of the table specified by the schema.
    """

    def __set_name__(self, owner: Type[Instance], name: str) -> None:
        self._name = name

    def __get__(self, instance: Instance, owner: Type[Instance]) -> Column:
        try:
            return instance[getattr(instance.schema, self._name)]
        except (KeyError, ValueError):
            raise ValueError(f"Schema has no {self._name} column identified.")


class CoordinateColumn(SchemaColumn):
    def __init__(self, latlon: str, frame: GenericFrame):
        self.latlon: str = latlon
        self.frame: GenericFrame = frame

    def __get__(
        self, instance: Instance, owner: Type[Instance]
    ) -> Latitude | Longitude:
        positions = instance.get_coordinates()
        positions = positions.transform_to(self.frame)

        return getattr(positions.frame.spherical, self.latlon)


class SourceTable(Table):
    """
    ``pyXMIP`` specific form of :py:class:`astropy.table.table.Table` for representing source catalogs and interacting
    with :py:class:`schema.SourceTableSchema`.
    """

    _schema: SourceTableSchema = TableAttribute()

    # =========================================== #
    # Columns                                     #
    # =========================================== #
    # Coordinate Columns #
    RA: Column = SchemaColumn()
    """The column containing the object RA coordinate."""
    DEC: Column = SchemaColumn()
    """The column containing the object DEC coordinate."""
    RA_ERR: Column = SchemaColumn()
    """The column containing the object RA coordinate error."""
    DEC_ERR: Column = SchemaColumn()
    """The column containing the object DEC coordinate error."""
    GAL_L: Column = SchemaColumn()
    """The column containing the object's galactic longitude."""
    GAL_B: Column = SchemaColumn()
    """The column containing the object's galactic latitude."""
    GAL_L_ERR: Column = SchemaColumn()
    """The column containing the object's galactic longitude error."""
    GAL_B_ERR: Column = SchemaColumn()
    """The column containing the object's galactic latitude error."""
    SEP: Column = SchemaColumn()
    """The column containing the separation between the search position and the output object."""
    # Name columns #
    NAME: Column = SchemaColumn()
    """The column containing the object's main identifier."""
    # Extra Columns #
    Z: Column = SchemaColumn()
    """The column containing the object's redshift."""
    TYPE: Column = SchemaColumn()
    """The column containing the object type of the catalog object."""

    @property
    def schema(self) -> SourceTableSchema:
        """
        The :py:class:`schema.SourceTableSchema` associated with this table.
        """
        if self._schema is None:
            self.generate_schema()
        return self._schema

    @schema.setter
    def schema(self, value: SourceTableSchema) -> None:
        self._schema = value

    @property
    def lon(self) -> units.Quantity:
        """The native longitude (default coordinate frame) of the catalog.

        Notes
        -----

        This property accesses the :py:attr:`schema.SourceTableSchema.coordinate_columns` attribute to determine what the
        default coordinate columns are. It then accesses them and returns the corresponding column of the table.

        If the column has a ``.unit`` attribute, that will be utilized; otherwise, the column is assumed to have the
        default schema units.

        See Also
        --------

        lat, lon, get_coordinates

        """
        unit = self.schema.coordinate_columns[0].unit
        if unit is None:
            unit = self.schema.default_angle_units

        return enforce_units(
            units.Quantity(self[self.schema.coordinate_columns[0].name]),
            unit,
        )

    @property
    def lat(self) -> units.Quantity:
        """The native latitude (default coordinate frame) of the catalog.

        Notes
        -----

        This property accesses the :py:attr:`schema.SourceTableSchema.coordinate_columns` attribute to determine what the
        default coordinate columns are. It then accesses them and returns the corresponding column of the table.

        If the column has a ``.unit`` attribute, that will be utilized; otherwise, the column is assumed to have the
        default schema units.

        See Also
        --------

        lat, lon, get_coordinates

        """
        unit = self.schema.coordinate_columns[1].unit
        if unit is None:
            unit = self.schema.default_angle_units

        return enforce_units(
            units.Quantity(self[self.schema.coordinate_columns[1].name]),
            unit,
        )

    def get_coordinates(self) -> SkyCoord:
        """
        Obtain the default coordinate positions for the catalog objects.

        Returns
        -------
        :py:class:`astropy.coordinates.SkyCoord`
            The coordinates of the sources in default frame.

        See Also
        --------

        lat, lon, get_coordinates
        """
        return SkyCoord(self.lon, self.lat, frame=self.schema.coordinate_frame)

    def generate_schema(self) -> None:
        """
        Generate and assign a default / automated schema to this table.

        Notes
        -----

        See :py:meth:`schema.SourceTableSchema.construct`

        """
        self._schema = SourceTableSchema.from_table(self)

    def get_formatted_types(self):
        """
        Returns an altered version of the :py:attr:`SourceTable.TYPE` column with standardized formatting.

        Returns
        -------
        Column
            The corrected formatting.

        Notes
        -----

        What this method does is detect the :py:attr:`schema.SourceTableSchema.object_type_separator` in each row,
        then reformat so that all of the types have the same standard formatting.

        """
        return format_table_types(self, self.schema)

    def count_types(self) -> Table:
        """
        Count the number of each object type in the catalog.

        Returns
        -------
        Table
            The count table. Formatted as the type followed by the number of such types.
        """
        assert (
            self.schema.TYPE in self.columns
        ), f"Cannot count types because there is no TYPE column {self.schema.TYPE}."

        # ------------------------------------------------- #
        # Correcting the formatting
        # ------------------------------------------------- #
        if len(self) == 0:
            # This is operational redundancy. If the table has no entries, we can just return 0 for each object type.
            return Table({k: [0] for k in self.schema.object_map})

        # Collect the separator #
        _sep = self.schema.object_type_separator

        # -- Managing types for non-zero length -- #
        _types = pd.Series(
            self.get_formatted_types()
        )  # --> this does all the reformatting.
        return Table(
            {
                k: [
                    len(_types[_types.str.contains(f"{re.escape(f'{_sep}{k}{_sep}')}")])
                ]
                for k in self.schema.object_map
            }
        )

    def append_to_fits(self, path, hdu):
        """
        Append this table to an existing fits table.

        Parameters
        ----------
        path: str
            The path to the fits file to append to.
        hdu: str
            The name of the HDU to append to.

        """
        _self_hdu = fits.table_to_hdu(self)

        try:
            with fits.open(path, "update") as hdu_list:
                if hdu in hdu_list:
                    _hdul, _len_hdul = hdu_list[hdu], len(hdu_list[hdu].data)
                    new_hudl = fits.BinTableHDU.from_columns(
                        _hdul.columns, nrows=_len_hdul + len(self)
                    )
                    for colname in hdu_list[hdu].columns.names:
                        new_hudl.data[colname][_len_hdul:] = _self_hdu.data[colname]

                    del hdu_list[hdu]
                else:
                    new_hudl = fits.table_to_hdu(self)

                new_hudl.name = hdu
                hdu_list.append(new_hudl)
                hdu_list.flush()
        except Warning:
            raise ValueError()

    def append_to_sql(self, table, conn, **kwargs):
        self.to_pandas().to_sql(
            str(table).upper(), con=conn, if_exists="append", index=False, **kwargs
        )

    @classmethod
    def _convert_types(cls, table, schema):
        # -- omap -- #
        omap = schema.object_map

        # -- pull the table's types -- #
        _sep = schema.object_type_separator
        _types = pd.Series(format_table_types(table, schema))
        _types = _types.str[1:-1].str.split(_sep)  # --> convert to the string subsets.

        # -- Fixing the table types -- #
        _types = _types.apply(
            lambda x: "|" + "|".join([omap.get(l, f"NSM_{l}") for l in x]) + "|"
        )

        table[schema.TYPE] = _types

        return table


def format_table_types(table: Table, schema: SourceTableSchema = None) -> list:
    """
    Returns an altered version of the :py:attr:`SourceTable.TYPE` column with standardized formatting.

    Parameters
    ----------
    table: Table
        The table from which the types should be formatted.
    schema: SourceTableSchema, optional
        The :py:class:`schema.SourceTableSchema` associated with the table. If ``table`` is a :py:class:`SourceTable`, then
        this will be determined automatically. Otherwise, the schema must be provied.

    Returns
    -------
    Column
        The corrected formatting.

    Notes
    -----

    What this method does is detect the :py:attr:`schema.SourceTableSchema.object_type_separator` in each row,
    then reformat so that all of the types have the same standard formatting.

    """
    if schema is None:
        if isinstance(table, SourceTable):
            schema = table.schema
        else:
            raise ValueError(
                "The table you provided doesn't have a built-in schema. One must be provided explicitly."
            )

    # grab the separator for the type column
    _sep = schema.object_type_separator

    # iterate through and correct.
    _types = list(table[schema.TYPE])

    for i, val in enumerate(_types):
        if not len(val):
            # This is a non-specified type. We set it to the ? type.
            _types[i] = f"{_sep}?{_sep}"
            continue

        if val[0] != _sep:
            # Correct the first column to enforce the separator.
            _types[i] = _sep + _types[i]

        if val[-1] != _sep:
            # Correct the final column.
            _types[i] = _types[i] + _sep

    return _types


def load(path, *args, **kwargs):
    """
    Load a catalog into ``pyXMIP``.

    Parameters
    ----------
    path: str
        The path to the catalog to load.


    Returns
    -------
    :py:class:`SourceTable`
        The resulting table.
    """
    return SourceTable.read(path, *args, **kwargs)


def correct_column_types(table):
    # -- Fix object column types -- #
    for col in table.columns:
        if table[col].dtype == "object":
            table[col] = np.array([str(j) for j in table[col]], dtype="<U64")

    return table



import os
import yaml
from pyXMIP.utilities.core import bin_directory
from pyXMIP.utilities.logging import mainlog
from pyXMIP._script_commands import CLI_FUNCTIONS,CLIParser
# ================================================================================= #
# Reading the yaml file
# ================================================================================= #


scripts_path = os.path.join(bin_directory, "scripts.yaml")

with open(scripts_path, "r") as file:
    _yaml_cli = yaml.load(file, yaml.FullLoader)


# ================================================================================= #
# CLI Configuration
# ================================================================================= #
# Setup the main parser and add the commands subparser.
_main_parser = CLIParser.build_recursive(_yaml_cli)

# ================================================================================ #
# RUN
# ================================================================================ #
args = dict(_main_parser.parse_args().__dict__)

operation = args.pop("operation", None)
if operation is None:
    from pyXMIP.utilities.text import print_cli_header

    print_cli_header()
    mainlog.error("Please enter a sub-command.")
    exit(1)
elif operation == "unimplemented":
    from pyXMIP.utilities.text import print_cli_header

    print_cli_header()
    mainlog.error("This is not yet implemented.")
    exit(1)

else:
    CLI_FUNCTIONS[operation](**args)
    exit(0)


"""
Schema management module for ``pyXMIP``

The :py:mod:`schema` module provides structure and data validation processes for user specified objects. In most cases,
this functions to effectively allow ``pyXMIP`` to convert from the user's data convention to those used internally.

Notes
-----

For a long-form discussion about schema, see :ref:`schema`.

"""
import os
import pathlib as pt
from typing import ClassVar, Type, Union

import astropy.coordinates as astro_coords
import pydantic as pyn

from pyXMIP.utilities.core import bin_directory, find_descriptors, rgetattr, rsetattr
from pyXMIP.utilities.logging import mainlog
from pyXMIP.utilities.types import (
    ColumnMap,
    ICRSCoordinateStdErrorSpecifier,
    Registry,
    SourceTableSchemaSettings,
    TableColumn,
    construct_template,
)

#: directory containing the built-in schema.
builtin_schema_directory = os.path.join(bin_directory, "builtin_schema")
#: directory containing the built-in :py:class:`SourceTableSchema` instances.
builtin_source_table_schema_directory = os.path.join(
    builtin_schema_directory, "source_table"
)
#: directory containing the built-in :py:class:`ReductionSchema` instances.
builtin_reduction_schema_directory = os.path.join(builtin_schema_directory, "reduction")


class SchemaEntry:
    """
    Descriptor class for accessing variables in the :py:class:`Schema` class.
    """

    def __init__(self, dict_location: str = None):
        """
        Initialize the entry.

        Parameters
        ----------
        dict_location: str
            The position of the underlying attribute in the :py:class:`Schema` instance. For example,
            the redshift column is ``'schema.column_map.Z'``, so ``dict_location = 'column_map'``.
        """
        self.dict_location = dict_location

    def __set_name__(self, owner, name):
        self._name = name

        if self.dict_location is None:
            self.dict_location = self._name
        else:
            self.dict_location += f".{self._name}"

    def __get__(self, instance, owner):
        try:
            return rgetattr(instance, self.dict_location)
        except AttributeError:
            return None

    def __set__(self, instance, value):
        rsetattr(instance, self.dict_location, value)


class ColumnEntry:
    """
    Specialized descriptor for accessing :py:class:`Schema` columns.
    """

    def __init__(self, auto_id=None, default_unit=None):
        self.auto_id = auto_id if auto_id is not None else []
        self.default_unit = default_unit

    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        try:
            return getattr(instance.column_map, self._name).name
        except AttributeError:
            return None

    def __set__(self, instance, value):
        if isinstance(value, TableColumn):
            setattr(instance.column_map, self._name, value)
        elif isinstance(value, dict):
            setattr(instance.column_map, self._name, TableColumn(**value))
        else:
            setattr(instance.column_map, self._name, TableColumn(name=value))


class Schema(pyn.BaseModel):
    """
    The generic schema class from which all other ``pyXMIP`` schemas are derived.

    .. warning::

        This is an abstract class.

    Notes
    -----

    The :py:class:`Schema` class is simply a wrapper on the PyDantic :py:class:`pydantic.BaseModel` class. This allows
    for the validation of user provided data.

    """

    def __init__(self, **mapping):
        """
        Initialize the generalized schema class.

        Parameters
        ----------
        mapping:
            Keyword arguments to pass into the schema.
        """
        super().__init__(**mapping)

    def __repr__(self):
        return f"<{self.__class__.__name__} instance>"

    def write(
        self,
        filename: Union[str, pt.Path],
        overwrite: bool = False,
        file_format: str = "yaml",
    ):
        """
        Write this schema instance to file. Various data formats are permitted.

        Parameters
        ----------
        filename:str
            The path at which to write the schema.
        overwrite: bool, optional
            If ``True``, then write will occur regardless of the prior existence of the file. Otherwise, an error is
            raised.
        file_format: str, optional
            The format to select for outputting the model.

        Raises
        ------
        IOError
            Raised if the specified ``filename`` cannot be reached.
        AssertionError
            Raised if the file already exists and ``overwrite=False``.

        """
        # ----------------------------------------------------- #
        # Arg / kwarg validation                                #
        # ----------------------------------------------------- #
        path = pt.Path(filename)

        # check the path for validity. The upstream directory must already exist.
        if not path.parents[0].exists():
            raise IOError(f"Parent directory(ies) of {path} don't exist.")

        # Check if the file already exists and has data.
        if not overwrite:
            # We don't have overwrite. Check if we can proceed.
            assert (
                not path.exists()
            ), f"Cannot write to {filename} because it already exists and overwrite is False."
        elif overwrite and path.exists():
            # We have overwrite = True and the file exists, so we delete it.
            path.unlink()

        # ----------------------------------------------------- #
        # Formatting Determination                              #
        # ----------------------------------------------------- #
        match file_format:  # Check the various formats.
            case "yaml":
                # This will be output to yaml.
                import yaml

                _model_dict = self.dict()

                with open(filename, "w") as f:
                    yaml.dump(_model_dict, f)

            case "json":
                import json

                _model_str = self.model_dump(mode="json")

                with open(filename, "w") as f:
                    json.dump(_model_str, f)

            case "toml":
                import toml

                _model_dict = self.dict()

                with open(filename, "w") as f:
                    toml.dump(_model_dict, f)
            case "txt":
                _model_str = str(self.dict())

                with open(filename, "w") as f:
                    f.write(_model_str)

    @classmethod
    def read(
        cls, filename: Union[str, pt.Path], file_format: str = None, *args, **kwargs
    ):
        """
        Read schema from file. Various data formats are permitted.

        Parameters
        ----------
        filename:str
            The path at which to read the schema.
        file_format: str, optional
            The intended format for the schema being read. If ``None``, then ``pyXMIP`` will attempt
            to determine it based on the suffix.

        Raises
        ------
        IOError
            Raised if the specified ``filename`` cannot be reached.
        AssertionError
            Raised if the file already exists and ``overwrite=False``.

        """
        # ----------------------------------------------------- #
        # Arg / kwarg validation                                #
        # ----------------------------------------------------- #
        path = pt.Path(filename)

        # check the path for validity. The upstream directory must already exist.
        if not path.exists():
            raise IOError(f"{path} doesn't exist.")

        # ----------------------------------------------------- #
        # Formatting Determination                              #
        # ----------------------------------------------------- #
        if file_format is None:
            file_format = path.suffix[1:]

        match file_format:  # Check the various formats.
            case "yaml":
                # This will be output to yaml.
                import yaml

                with open(filename, "r") as f:
                    if not len(args):
                        # the args should have specified the loader. Because they're not provided, we assume full loader.
                        args = [yaml.FullLoader]

                    _r = yaml.load(f, *args, **kwargs)

            case "json":
                import json

                with open(filename, "r") as f:
                    _r = json.load(f, *args, **kwargs)
            case "toml":
                import toml

                with open(filename, "r") as f:
                    _r = toml.load(f, *args, **kwargs)

            case "txt":
                with open(filename, "r") as f:
                    _r = {k: v for k, v in [u.split(":") for u in f.read().split(",")]}
            case _:
                raise ValueError(f"Failed to recognize format {file_format}.")

        return cls(**_r)


class SourceTableSchema(Schema):
    """
    Schema for interactions with :py:class:`structures.table.SourceTable`.
    """

    # =========================================== #
    # Validated parameters                        #
    # =========================================== #
    column_map: ColumnMap
    """:py:class:`utilities.types.ColumnMap`: Mapping between table columns and standard recognized by ``pyXMIP``.

    The column map instructs ``pyXMIP`` about what different columns in a table mean.
    """
    object_map: dict | None = None
    """dict: The mapping between the native source object types and the standardized types.

    This parameter is used to determine how to map the typing from the catalog / table into the standardized
    SIMBAD types used by ``pyXMIP``.

    .. hint::

        You can find a complete list of SIMBAD object types `here <https://simbad.cds.unistra.fr/guide/otypes.htx>`_.
    """
    settings: SourceTableSchemaSettings = SourceTableSchemaSettings()
    """:py:class:`utilities.types.SourceTableSchemaSettings`: The schema settings for this instance."""

    # =========================================== #
    # Columns                                     #
    # =========================================== #
    # Coordinate Columns #
    RA: ClassVar[str] = ColumnEntry(auto_id=["RA", "ra", "Ra"], default_unit="deg")
    """The column containing the object RA coordinate."""
    DEC: ClassVar[str] = ColumnEntry(auto_id=["dec", "DEC", "Dec"], default_unit="deg")
    """The column containing the object DEC coordinate."""
    RA_ERR: ClassVar[str] = ColumnEntry(auto_id=[], default_unit="arcsec")
    """The column containing the object RA coordinate error."""
    DEC_ERR: ClassVar[str] = ColumnEntry(auto_id=[], default_unit="arcsec")
    """The column containing the object DEC coordinate error."""
    GAL_L: ClassVar[str] = ColumnEntry(
        auto_id=["Gal_L", "GL", "LII"], default_unit="deg"
    )
    """The column containing the object's galactic longitude."""
    GAL_B: ClassVar[str] = ColumnEntry(
        auto_id=["GAL_B", "GB", "BII"], default_unit="deg"
    )
    """The column containing the object's galactic latitude."""
    GAL_L_ERR: ClassVar[str] = ColumnEntry(auto_id=[], default_unit="arcsec")
    """The column containing the object's galactic longitude error."""
    GAL_B_ERR: ClassVar[str] = ColumnEntry(auto_id=[], default_unit="arcsec")
    """The column containing the object's galactic latitude error."""
    SEP: ClassVar[str] = ColumnEntry(auto_id=["Separation"], default_unit="arcsec")
    # Name columns #
    NAME: ClassVar[str] = ColumnEntry(
        auto_id=["IAUNAME", "NAME", "MAIN_ID", "OBJECT_ID"]
    )
    """The column containing the object's main identifier."""
    # Extra Columns #
    Z: ClassVar[str] = ColumnEntry(auto_id=["REDSHIFT", "redshift", "Z", "z"])
    """The column containing the object's redshift."""
    TYPE: ClassVar[str] = ColumnEntry(auto_id=["TYPE"])
    """The column containing the object type of the catalog object."""

    # ============================================ #
    # Settings                                     #
    # ============================================ #
    default_coord_system: ClassVar[str] = SchemaEntry(dict_location="settings")
    """str: The default coordinate system to use.

    Use this to select the native coordinate system for the schema. Whichever coordinate system is selected, there must
    be sufficient identifiable columns to actually obtain those coordinates.

    """
    object_type_separator: ClassVar[str] = SchemaEntry(dict_location="settings")
    """str: The separator character / string between different object types.

    In many databases, returned object types may have multiple types separated by a delimiter. This delimiter should
    be specified here if it is needed.

    .. important::

        In all native ``pyXMIP`` data objects, a "|" is used for the delimiter. Even if the object type is a single
        value, all object types in ``pyXMIP`` tables are formatted ``"|entry_1|entry_2|...|"``.

    """

    _coordinate_requirements: ClassVar[dict] = {
        astro_coords.ICRS: [RA, DEC],
        astro_coords.Galactic: [GAL_L, GAL_B],
    }

    def available_coordinate_frames(self):
        """
        List the available astropy coordinate frames which can be used to determine positions of the data based on
        columns provided to this schema.

        Returns
        -------
        dict
            The available coordinate frames.
        """
        # -- determine which coordinate frames are available -- #
        available_frames = [
            frame
            for frame, cols in self._coordinate_requirements.items()
            if all(getattr(self, col._name) is not None for col in cols)
        ]

        return {
            frame: [u._name for u in self._coordinate_requirements[frame]]
            for frame in available_frames
        }

    @property
    def coordinate_system(self) -> tuple[astro_coords.GenericFrame, list[str]]:
        """
        The coordinate system and associated columns for this schema. Unless this attribute is set after initialization, the
        coordinate system will be the same as determined by :py:attr:`SourceTableSchema.default_coord_system`.

        Returns
        -------
        :py:class:`astropy.coordinates.GenericFrame`
            The correct coordinate frame object for the coordinate system used by the schema.
        list
            The table column names corresponding to the coordinate system.
        """
        _cs = getattr(astro_coords, self.default_coord_system)
        return _cs, self.available_coordinate_frames()[_cs]

    @property
    def coordinate_frame(self) -> astro_coords.GenericFrame:
        """The specific coordinate system currently corresponding to this schema."""
        return self.coordinate_system[0]

    @property
    def coordinate_columns(self) -> list[TableColumn]:
        """The columns representing the current coordinate system."""
        # The setup returns the actual descriptor object, we need the underlying parameter (hence the call to .__get__)
        return [
            getattr(self.column_map, coord_col)
            for coord_col in self.coordinate_system[1]
        ]

    @coordinate_system.setter
    def coordinate_system(self, value: Union[str, astro_coords.GenericFrame]):
        # check the typing.
        if isinstance(value, str):
            pass
        elif isinstance(value, astro_coords.GenericFrame):
            value = value.__name__
        else:
            raise TypeError(f"Cannot set coordinate system to type {type(value)}.")

        # check validity
        assert hasattr(
            astro_coords, value
        ), f"The coordinate frame {value} is not in astropy's coordinate systems."
        assert (
            getattr(astro_coords, value) in self.available_coordinate_frames()
        ), f"The coordinate frame {value} is not a valid coordinate frame for this schema."

        self.settings.default_coord_system = value

    @classmethod
    def from_table(cls, table):
        r"""
        Construct a "best-guess" schema from a table by deducing the meaning of its columns.

        Parameters
        ----------
        table: :py:class:`structures.table.SourceTable` or :py:class:`pd.DataFrame`
            The table on which to determine the best guess schema.

        Returns
        -------
        :py:class:`SourceTableSchema`
            The finalized guess schema.
        """
        mainlog.debug(f" [SourceTableSchema] Constructing {cls.__name__} from table.")

        # ----------------------------------------------#
        # setup components
        _table_columns = table.columns  # isomorphic call between table types.
        _constructed_schema = construct_template(cls)

        _default_coordinate_frame = None

        # ----------------------------------------------------------------------#
        # Managing special columns
        # ----------------------------------------------------------------------#
        _constructed_schema["column_map"] = {}
        for k, v in find_descriptors(cls, ColumnEntry).items():
            # search for matching keys
            _matched_keys = [key for key in _table_columns if key in v.auto_id]

            if len(_matched_keys):
                # A matching key was found.
                _constructed_schema["column_map"][k] = {"name": _matched_keys[0]}

                # try to fetch a unit.
                if hasattr(table[_matched_keys[0]], "unit"):
                    _constructed_schema["column_map"][k]["unit"] = table[
                        _matched_keys[0]
                    ].unit
                else:
                    _constructed_schema["column_map"][k]["unit"] = v.default_unit

                if _constructed_schema["column_map"][k]["unit"] is None:
                    del _constructed_schema["column_map"][k]["unit"]

                mainlog.debug(
                    f" [SourceTableSchema] Identified special column {k} with column {_matched_keys[0]} [unit={_constructed_schema['column_map'][k].get('unit','')}] of the table."
                )

            else:
                mainlog.debug(
                    f" [SourceTableSchema] Failed to identify automatic match for special column {k}."
                )

        # ----------------------------------------------------------------------#
        # Managing object maps
        # ----------------------------------------------------------------------#
        _constructed_schema["object_map"] = {}

        # ----------------------------------------------------------------------#
        # Managing coordinates
        # ----------------------------------------------------------------------#
        _construction_coords_avail = [
            frame
            for frame, cols in cls._coordinate_requirements.items()
            if all(col._name in _constructed_schema["column_map"] for col in cols)
        ]
        assert len(
            _construction_coords_avail
        ), " [SourceTableSchema] No valid coordinate system could be determined."

        _default_coordinate_frame = _construction_coords_avail[0].__name__
        _constructed_schema["settings"].default_coord_system = _default_coordinate_frame

        mainlog.debug(
            f" [SourceTableSchema] Located {len(_construction_coords_avail)} possible coordinate frames. Selected {_default_coordinate_frame} as default."
        )
        return cls(**_constructed_schema)

    def guess_coordinate_standard_error_struct(self):
        # -- build the base template -- #
        template = construct_template(ICRSCoordinateStdErrorSpecifier)

        # --------------------------- #
        # Look for valid coordinates  #
        # --------------------------- #
        _available_coordinate_frames = self.available_coordinate_frames()

        # For each of the available frames, we need to determine if the coordinate errors are available.
        _frames_with_error_cols = {}
        for frame, required_columns in _available_coordinate_frames.items():
            required_error_columns = [
                getattr(self.column_map, u + "_ERR")
                if hasattr(self.column_map, u + "_ERR")
                else None
                for u in required_columns
            ]
            required_columns = [
                getattr(self.column_map, u) for u in required_columns
            ]  # Warning! note the change in type.

            if any(u is None for u in required_error_columns):
                continue
            else:
                _frames_with_error_cols[frame] = (
                    required_error_columns,
                    required_columns,
                )

        if not len(_frames_with_error_cols):
            # We can't find a way to improve on the base template. Return it blank.
            return template
        else:
            if self.coordinate_system[0] in _frames_with_error_cols:
                err_cols = (
                    self.coordinate_system[0],
                    _frames_with_error_cols[self.coordinate_system[0]],
                )
            else:
                err_cols = [(k, v) for k, v in _frames_with_error_cols.items()][0]

        # ------------------------------------ #
        # Adding error columns to the template #
        # ------------------------------------ #
        template["longitude_column"], template["latitude_column"] = (
            err_cols[1][1][0].dict(),
            err_cols[1][1][1].dict(),
        )
        (
            template["longitude_column_error"]["base_column"],
            template["latitude_column_error"]["base_column"],
        ) = (err_cols[1][1][0].dict(), err_cols[1][1][1].dict())
        (
            template["longitude_column_error"]["error_column"],
            template["latitude_column_error"]["error_column"],
        ) = (err_cols[1][0][0].dict(), err_cols[1][0][1].dict())

        return template


class CMDSchema(Schema):
    """
    Schema class for interaction with :py:class:`cross_reference.CrossMatchDatabase`.

    Notes
    -----

    This class is effectively a collection of :py:class:`SourceTableSchema` instances compiled together to
    represent the entire :py:class:`cross_reference.CrossMatchDatabase` structure.

    """

    table_schema: dict[str, SourceTableSchema]
    """dict: Dictionary of schema specific to each of the cross matching databases.

    For each of the tables in the underlying ``SQL`` database (excluding ``CATALOG``), there should be a schema corresponding
    to the table format for the source. For the most part, these are the :py:attr:`structures.databases.RemoteDatabase.default_query_schema` of
    the queried databases.

    .. hint::

        For the most part, these are generated automatically when the cross matching occurs and you don't need to worry about
        generating one of these on your own. If you do, the dictionary has a simple format: keys are the NAMES of the tables in
        the :py:class:`cross_reference.CrossMatchDatabase`, the values should be the corresponding schema.
    """
    catalog_schema: SourceTableSchema
    """:py:class:`SourceTableSchema` The schema for the catalog.

    This is the schema corresponding to the original catalog. In almost all cases, this is done automatically.
    """


class ReductionSchema(Schema):
    pass


class SchemaRegistry(Registry):
    """
    Registry class for containing collections of :py:class:`Schema` classes.
    """

    def __init__(self, mapping: dict[str, str | pt.Path], schema_class: Type[Schema]):
        """
        Initialize the :py:class:`SchemaRegistry`

        Parameters
        ----------
        mapping: dict
            The dictionary of ``name``,``path`` pairs to use as the registry.
        schema_class:
            The schema class to initialize objects with.
        """
        super().__init__(mapping)

        self.schema_class: Type[Schema] = schema_class

    def __getitem__(self, item: str) -> Schema:
        return self.schema_class.read(super().__getitem__(item))

    @classmethod
    def from_directory(cls, directory, schema_class: Type[Schema]):
        """
        Generate a registry of :py:class:`Schema` instances from a directory of files.

        Parameters
        ----------
        directory: str
            The directory to load from.
        schema_class:
            The schema class for this registry.

        Returns
        -------
        :py:class:`SchemaRegistry`
            The corresponding registry.
        """
        import pathlib as pt

        directory = pt.Path(directory)
        assert directory.exists(), f"The directory {directory} doesn't exist."
        assert directory.is_dir(), f"The path {directory} isn't a directory."

        sub_objects = os.listdir(directory)
        _dict = {}

        for o in sub_objects:
            pth = pt.Path(os.path.join(directory, o))
            if pth.is_file():
                suffix = pth.suffix

                if suffix in [".toml", ".json", ".yaml"]:
                    _dict[pth.name.replace(pth.suffix, "")] = str(pth)

        return cls(_dict, schema_class)


#: The default :py:class:`SourceTableSchema` registry.
DEFAULT_SOURCE_SCHEMA_REGISTRY = SchemaRegistry.from_directory(
    builtin_source_table_schema_directory, SourceTableSchema
)

#: The default :py:class:`ReductionSchema` registry.
DEFAULT_REDUCTION_SCHEMA_REGISTRY = SchemaRegistry.from_directory(
    builtin_reduction_schema_directory, ReductionSchema
)


if __name__ == "__main__":
    NED_schema = DEFAULT_SOURCE_SCHEMA_REGISTRY["NED"]
    print(type(NED_schema.model_dump(mode="json")))



"""
Gaussian Process regression tools.
"""
import numpy as np
from sklearn.gaussian_process.kernels import (
    Hyperparameter,
    Kernel,
    NormalizedKernelMixin,
    StationaryKernelMixin,
    _check_length_scale,
)
from sklearn.metrics.pairwise import haversine_distances


class Haversine_RBF(StationaryKernelMixin, NormalizedKernelMixin, Kernel):
    """Radial basis function kernel (aka squared-exponential kernel) with a haversine distance metric to
    allow for the use of spherical data.

    The RBF kernel is a stationary kernel. It is also known as the
    "squared exponential" kernel. It is parameterized by a length scale
    parameter :math:`l>0`, which can either be a scalar (isotropic variant
    of the kernel) or a vector with the same number of dimensions as the inputs
    X (anisotropic variant of the kernel). The kernel is given by:

    .. math::
        k(x_i, x_j) = \\exp\\left(- \\frac{d(x_i, x_j)^2}{2l^2} \\right)

    where :math:`l` is the length scale of the kernel and
    :math:`d(\\cdot,\\cdot)` is the Euclidean distance.
    For advice on how to set the length scale parameter, see e.g. [1]_.

    This kernel is infinitely differentiable, which implies that GPs with this
    kernel as covariance function have mean square derivatives of all orders,
    and are thus very smooth.
    See [2]_, Chapter 4, Section 4.2, for further details of the RBF kernel.

    Read more in the :ref:`User Guide <gp_kernels>`.

    .. versionadded:: 0.18

    Parameters
    ----------
    length_scale : float or ndarray of shape (n_features,), default=1.0
        The length scale of the kernel. If a float, an isotropic kernel is
        used. If an array, an anisotropic kernel is used where each dimension
        of l defines the length-scale of the respective feature dimension.

    length_scale_bounds : pair of floats >= 0 or "fixed", default=(1e-5, 1e5)
        The lower and upper bound on 'length_scale'.
        If set to "fixed", 'length_scale' cannot be changed during
        hyperparameter tuning.

    References
    ----------
    .. [1] `David Duvenaud (2014). "The Kernel Cookbook:
        Advice on Covariance functions".
        <https://www.cs.toronto.edu/~duvenaud/cookbook/>`_

    .. [2] `Carl Edward Rasmussen, Christopher K. I. Williams (2006).
        "Gaussian Processes for Machine Learning". The MIT Press.
        <http://www.gaussianprocess.org/gpml/>`_

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.gaussian_process import GaussianProcessClassifier
    >>> from sklearn.gaussian_process.kernels import RBF
    >>> X, y = load_iris(return_X_y=True)
    >>> kernel = 1.0 * Haversine_RBF(1.0)
    >>> gpc = GaussianProcessClassifier(kernel=kernel,
    ...         random_state=0).fit(X, y)
    >>> gpc.score(X, y)
    0.9866666666666667
    >>> gpc.predict_proba(X[:2,:])
    array([[0.83548752, 0.03228706, 0.13222543],
           [0.79064206, 0.06525643, 0.14410151]])
    """

    def __init__(self, length_scale=1.0, length_scale_bounds=(1e-5, 1e5)):
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds

    @property
    def anisotropic(self):
        return np.iterable(self.length_scale) and len(self.length_scale) > 1

    @property
    def hyperparameter_length_scale(self):
        if self.anisotropic:
            return Hyperparameter(
                "length_scale",
                "numeric",
                self.length_scale_bounds,
                len(self.length_scale),
            )
        return Hyperparameter("length_scale", "numeric", self.length_scale_bounds)

    def __call__(self, X, Y=None, eval_gradient=False):
        """Return the kernel k(X, Y) and optionally its gradient.

        Parameters
        ----------
        X : ndarray of shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)

        Y : ndarray of shape (n_samples_Y, n_features), default=None
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.

        eval_gradient : bool, default=False
            Determines whether the gradient with respect to the log of
            the kernel hyperparameter is computed.
            Only supported when Y is None.

        Returns
        -------
        K : ndarray of shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)

        K_gradient : ndarray of shape (n_samples_X, n_samples_X, n_dims), \
                optional
            The gradient of the kernel k(X, X) with respect to the log of the
            hyperparameter of the kernel. Only returned when `eval_gradient`
            is True.
        """
        X = np.atleast_2d(X)
        length_scale = _check_length_scale(X, self.length_scale)
        if Y is None:
            dists = haversine_distances(X) ** 2
            K = np.exp(-0.5 * dists / length_scale**2)
            # convert from upper-triangular matrix to square matrix
            # K = squareform(K)
            np.fill_diagonal(K, 1)
        else:
            if eval_gradient:
                raise ValueError("Gradient can only be evaluated when Y is None.")
            dists = haversine_distances(X, Y) ** 2
            K = np.exp(-0.5 * dists / length_scale**2)

        if eval_gradient:
            if self.hyperparameter_length_scale.fixed:
                # Hyperparameter l kept fixed
                return K, np.empty((X.shape[0], X.shape[0], 0))
            elif not self.anisotropic or length_scale.shape[0] == 1:
                K_gradient = (K * dists)[:, :, np.newaxis]
                return K, K_gradient
            elif self.anisotropic:
                # We need to recompute the pairwise dimension-wise distances
                K_gradient = (X[:, np.newaxis, :] - X[np.newaxis, :, :]) ** 2 / (
                    length_scale**2
                )
                K_gradient *= K[..., np.newaxis]
                return K, K_gradient
        else:
            return K

    def __repr__(self):
        if self.anisotropic:
            return "{0}(length_scale=[{1}])".format(
                self.__class__.__name__,
                ", ".join(map("{0:.3g}".format, self.length_scale)),
            )
        else:  # isotropic
            return "{0}(length_scale={1:.3g})".format(
                self.__class__.__name__, np.ravel(self.length_scale)[0]
            )



r"""Module for managing conversion between different coordinate conventions.

Notes
-----
The utilities in this module are largely focused on managing the interface between HEALPix, astronomical coordinates,
and mathematical convention.

In pyXMIP, coordinates are specified in 1 of 3 systems:

- **Mathematical Convention**: In the mathematical convention, spherical coordinates are denoted :math:`(r,\phi,\theta)`,
  where :math:`\phi` is the azimuthal angle measured from the :math:`x`-axis counter-clockwise around positive :math:`z`.
  :math:`\theta` is the elevation angle, measured from the :math:`xy`-plane to :math:`\pm \pi/2`.
- **HEALPix Convention**: The HEALPix coordinate convention is marginally different from the others in that it measures the
  elevation angle from :math:`0\to \pi`. In this convention, the angle :math:`\omega` is the *co-latitude* and measures from the
  north pole. The longitude :math:`\phi` is measured from :math:`0\to 2\pi` going eastward.
- **Coordinate Convention**: This is the convention for all the astronomical coordinate systems. These obey longitude and
  latitude as it is typically defined with longitude increasing to the east from 0 to 360 degrees.
"""
import numpy as np

# ================================================================================= #
# COORDINATE MANAGEMENT                                                             #
# ================================================================================= #

sky_coordinate_systems = {
    "latlon": {"phi": (-np.pi, "east"), "theta": (-np.pi / 2, "north")},
    "healpix": {"phi": (-np.pi, "east"), "theta": (np.pi, "south")},
    "math_convention": {"phi": (0, "east"), "theta": (-np.pi / 2, "north")},
}


def _get_system_offsets(to_system, from_system, dp, dt):
    po, to = sky_coordinate_systems[to_system]["phi"][0] - (
        dp * sky_coordinate_systems[from_system]["phi"][0]
    ), sky_coordinate_systems[to_system]["theta"][0] - (
        dt * sky_coordinate_systems[from_system]["theta"][0]
    )
    return po, to


def _get_system_direction(to_system, from_system):
    pd = (
        1
        if sky_coordinate_systems[to_system]["phi"][1]
        == sky_coordinate_systems[from_system]["phi"][1]
        else -1
    )
    td = (
        1
        if sky_coordinate_systems[to_system]["theta"][1]
        == sky_coordinate_systems[from_system]["theta"][1]
        else -1
    )
    return pd, td


def convert_coordinates(phi, theta, from_system="healpix", to_system="latlon"):
    # -- Sanity Checks -- #
    assert (
        from_system in sky_coordinate_systems
    ), f"Failed to recognize from_system {from_system}."
    assert (
        to_system in sky_coordinate_systems
    ), f"Failed to recognize to_system {to_system}."

    # -- calculating offset differences -- #
    phi_dir, theta_dir = _get_system_direction(to_system, from_system)
    phi_offset, theta_offset = _get_system_offsets(
        to_system, from_system, phi_dir, theta_dir
    )
    return (phi_dir * phi) + phi_offset, (theta_dir * theta) + theta_offset


def convert_skycoord(skycoord, to_system):
    # skycoords always go straight to latlon
    phi, theta = (
        skycoord.frame.spherical.lon.wrap_at("180 deg").rad,
        skycoord.frame.spherical.lat.rad,
    )

    return convert_coordinates(phi, theta, from_system="latlon", to_system=to_system)



"""
Module for interactions with external and local databases.

Notes
-----

For more information on databases, see :ref:`databases`.

"""
import os
import threading
import time
import warnings
from abc import ABC, abstractmethod
from itertools import repeat
from typing import Any, Callable, Generic, Type, TypeVar

import numpy as np
import requests.exceptions
import sqlalchemy as sql
from astropy import units
from astropy.coordinates import Angle, SkyCoord
from astropy.table import Table, vstack
from astroquery.ipac.ned import Ned
from astroquery.simbad import Simbad
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from pyXMIP.schema import DEFAULT_SOURCE_SCHEMA_REGISTRY, SourceTableSchema
from pyXMIP.structures.map import PoissonAtlas
from pyXMIP.structures.table import SourceTable, correct_column_types
from pyXMIP.utilities.core import bin_directory, enforce_units
from pyXMIP.utilities.logging import mainlog
from pyXMIP.utilities.types import Registry, convert_np_type_to_sql

poisson_map_directory: str = os.path.join(bin_directory, "psn_maps")

# -- Typing Variables -- #
Instance = TypeVar("Instance")
Value = TypeVar("Value")
Attribute = TypeVar("Attribute")


class DatabaseError(Exception):
    """
    Collective error type for issues with database queries.
    """

    def __init__(self, message: str = None):
        self.message = message
        super().__init__(self.message)


class _DatabaseConfigSetting(Generic[Instance, Attribute, Value]):
    # Descriptor class for query configuration settings.
    # Stored in query_config.
    def __init__(self, default: Any = None):
        self.default = default

    def __set_name__(self, owner: Type[Instance], name: str) -> None:
        self._name = name

    def __get__(self, instance: Instance, owner: Type[Instance]) -> Any:
        if self._name in instance.query_config:
            return instance.query_config[self._name]
        else:
            return self.default

    def __set__(self, instance: Instance, value: Any) -> None:
        instance.query_config[self._name] = value


class SourceDatabase(ABC):
    """
    Abstract class representation of a database class. All other database types are subclasses of this class.


    Notes
    -----

    .. warning::

        This is an abstract class; there is no reason to instantiate this class.

    """

    # -- class default variables and methods -- #
    default_poisson_atlas_path: str = os.path.join(poisson_map_directory, "NONE")
    """str: The path to the default poisson atlas file.

    For remote databases, this is generally non-trivial. For local databases,this is typically None.
    """
    default_query_config: dict[str, Any] = None
    """dict: The configuration information for performing the query.

    Depending on the particular database, this may vary substantially."""
    default_query_schema: SourceTableSchema = None
    """:py:class:`schema.SourceTableSchema`: The schema for the returned table after a query.

    For local databases, this is automatically set to the base catalog. For remotes, it must generally be constructed.
    """
    _thread_lock = threading.Lock()

    def __init__(self, db_name: str, *args, **kwargs):
        """
        Initialize the :py:class:`SourceDatabase`.

        Parameters
        ----------
        db_name: str
            The name of the database instance.
        poisson_path: str
            The path to the Poisson atlas file.
        query_config: dict
            Configuration settings for the querying process.
        query_schema: :py:class:`schema.SourceTableSchema`
            The schema to associated with query outputs.
        correct_query_output: callable
            The callable for correcting formatting of the query output.
        """
        self.name: str = db_name
        """str: The name of this database instance.
        """
        self._poisson_atlas_path: str = kwargs.pop(
            "poisson_path", self.__class__.default_poisson_atlas_path
        )
        """str: The path to the Poisson Atlas file.

        By default, it is ``None``; but remote subclasses should have a built-in poisson path.
        """
        self.query_config: dict = kwargs.pop(
            "query_config", self.__class__.default_query_config
        )
        """dict: The query configuration options.

        By default, it is ``None``; this varies for subclasses.
        """
        self.query_schema: SourceTableSchema = kwargs.pop(
            "query_schema", self.__class__.default_query_schema
        )
        """:py:class:`schema.SourceTableSchema`: The schema for the returned table after a query.

        For local databases, this is automatically set to the base catalog. For remotes, it must generally be constructed.
        """
        self.correct_query_output: Callable[[SourceTable], SourceTable] = kwargs.pop(
            "correct_query_output", self.default_correct_query_output
        )
        """callable: The method for altering returned source tables to be writable to sql.

        By default, these are pre-set. See the relevant documentation.
        """

    def __str__(self):
        return f"<SourceDatabase {self.name}>"

    def __repr__(self):
        return self.__str__()

    def register(self, registry: Registry):
        """Add this instance to a DBRegistry."""
        if self.name not in registry:
            registry[self.name] = self

    # -- Querying -- #

    @abstractmethod
    def _query_radius(self, position: SkyCoord, radius: units.Quantity) -> SourceTable:
        """
        DEVELOPERS: This is the query radius that should be altered in sub-classes.
        """
        pass

    def query_radius(self, position: SkyCoord, radius: units.Quantity) -> SourceTable:
        """
        Query the remote database at the specified position and pull all sources within a given radius.

        Parameters
        ----------
        position: :py:class:`astropy.coordinates.SkyCoord`
            The position at which to query.
        radius: :py:class:`astropy.units.Quantity`
            The angular area about which to query.

        Returns
        -------
        :py:class:`structures.table.SourceTable`

        Notes
        -----

        .. admonition:: Dev-Note

            This method doesn't need to be altered when writing a custom database class. It simply provides a standardized
            wrapper for the query. The ``_query_radius`` method contains the code which varies from class to class.

        Examples
        --------

        **Query for NED objects within 1 arcmin of the galactic center**:

        .. code-block:: python

            >>> from pyXMIP import NED
            >>> from astropy.coordinates import SkyCoord
            >>> from astropy import units

            >>> database_instance = NED()
            >>> query = database_instance.query_radius(SkyCoord(0,0,unit='deg',frame='galactic'),radius=1*units.arcmin)

            >>> print(f"The first matching object is {query[0]['Object Name']}, Separation={query[0]['Separation']} arcmin.")
            The first matching object is 2MASS J17453276-2856166, Separation=0.981 arcmin.

        """

        radius = enforce_units(radius, units.arcmin)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mainlog.info(
                f"Querying lat={np.round(position.frame.spherical.lat.deg,decimals=3)}, lon={np.round(position.frame.spherical.lon.deg,decimals=3)} in {self.name}..."
            )
            output_table = self._query_radius(
                position,
                radius,
            )

        return output_table

    @classmethod
    def _default_correct_query_output(
        cls, table: SourceTable | Table, schema: SourceTableSchema = None
    ) -> SourceTable | Table:
        """
        The default query output correction callable.

        This is only used if :py:attr:`SourceTable.correct_query_output` is not specified during initialization.

        Parameters
        ----------
        table: :py:class:`astropy.table.table.Table`
            The table being corrected for formatting.
        schema: :py:class:`schema.SourceTableSchema`
            The relevant query schema.

        Returns
        -------
        :py:class:`astropy.table.table.Table`
            The output table.

        Notes
        -----
        By default, the only thing that this function does is correct column types.

        """
        # -- remove all table meta-data -- #
        table.meta = None
        table.schema = schema if schema is not None else cls.default_query_schema

        # -- standardize type columns -- #
        table = correct_column_types(table)

        return table

    def default_correct_query_output(
        self, table: SourceTable | Table
    ) -> SourceTable | Table:
        """
        The default query output correction callable.

        This is only used if :py:attr:`SourceTable.correct_query_output` is not specified during initialization.

        Parameters
        ----------
        table: :py:class:`astropy.table.table.Table`
            The table being corrected for formatting.

        Returns
        -------
        :py:class:`astropy.table.table.Table`
            The output table.

        Notes
        -----
        By default, the only thing that this function does is correct column types.

        """
        return self.__class__._default_correct_query_output(
            table, schema=self.query_schema
        )

    def count(
        self, positions: SkyCoord, radii: units.Quantity, parallel_kwargs: dict = None
    ):
        """
        Count the number of each object type in a set of position queries.

        Parameters
        ----------
        positions: list of SkyCoord
            A list of sky coordinates to query at.
        radii: units.Quantity
            An array of the same size as ``positions`` containing the sample radii.
        parallel_kwargs: dict
            Parameters for parallelization. For more information see :ref:`parallelization`.

            .. note::

                This method implements multi-threading in some subclasses.

        Returns
        -------
        Table
            A table containing counts for each of the object types and the positions of queries.
        """
        from pyXMIP.utilities.optimize import map_to_threads

        # -------------------------------------------#
        # Managing args / kwargs and paths
        # -------------------------------------------#
        radii = enforce_units(radii, units.arcmin)
        if radii.isscalar:
            radii = radii * np.ones(len(positions))
        # ------------------------------------------------ #
        # Running the queries through the database         #
        # ------------------------------------------------ #
        with logging_redirect_tqdm(loggers=[mainlog]), mainlog.fixed_verb(
            info=0
        ), warnings.catch_warnings():
            warnings.simplefilter("ignore")
            pbar = tqdm(zip(positions, radii), total=len(radii))
            results = map_to_threads(
                self._thread_pooled_count,
                positions,
                radii,
                repeat(pbar),
                threading_kw=parallel_kwargs,
            )

            output_table = correct_column_types(
                vstack([k for k in results if k is not None])
            )

        output_table["RA"] = [p.icrs.ra for p in positions]
        output_table["DEC"] = [p.icrs.dec for p in positions]
        output_table["RAD"] = radii
        output_table["TIME"] = time.asctime()
        return output_table

    def _thread_pooled_count(
        self, position: SkyCoord, radius: units.Quantity, progress_bar: Any
    ) -> Table:
        try:
            _output_var = self.query_radius(position, radius).count_types()
        except Exception as exception:
            mainlog.error(exception.__repr__())
            _output_var = None
        progress_bar.update()
        return _output_var

    def random_sample_count(
        self, points: int, radii: units.Quantity, parallel_kwargs: Any = None
    ) -> Table:
        """
        Count the number of instances of each object type at each of a number of random positions on the sky.

        Parameters
        ----------
        points: int
            The number of randomly sampled points to query.
        radii: units.Quantity
            The radii of the search area for each query.
        parallel_kwargs: dict
            Parameters for parallelization. For more information see :ref:`parallelization`.

            .. note::

                This method implements multi-threading in some subclasses.

        Returns
        -------
        Table
            Table of counts for each of the object types.
        """
        from pyXMIP.stats.utilities import uniform_sample_spherical

        mainlog.info(f"Querying for {points} random counts on {self.name}.")

        # -------------------------------------------#
        # Managing args / kwargs and paths
        # -------------------------------------------#
        radii = enforce_units(radii, units.arcmin)
        if radii.isscalar:
            radii = radii * np.ones(points)

        # -------------------------------------------#
        # Pull the random samples
        # -------------------------------------------#
        phi, theta = uniform_sample_spherical(points)
        theta = np.pi / 2 - theta
        positions = SkyCoord(phi, theta, frame="galactic", unit="rad")

        return self.count(positions, radii, parallel_kwargs=parallel_kwargs)

    # -- Poisson Atlas related methods -- #

    @property
    def poisson_atlas(self) -> PoissonAtlas:
        """
        The Poisson-atlas corresponding to this database.

        Returns
        -------
        :py:class:`structures.map.PoissonAtlas`
            The corresponding atlas.

        Notes
        -----
        If the user wishes to set the Poisson-atlas, they should set ``instance.poisson_atlas = path``, where the path
        directs to the file in which the atlas is stored.

        .. warning ::

            Generally, these are ``None`` by default; particularly for :py:class:`LocalDatabase` instances. For built-in
            :py:class:`RemoteDatabase` classes, there are pre-built Poisson atlases by default.
        """
        try:
            return PoissonAtlas(self._poisson_atlas_path)
        except FileNotFoundError:
            mainlog.warning(
                f"There is no Poisson atlas at {self._poisson_atlas_path}. Reverting to default."
            )
            self._poisson_atlas_path = self.__class__.default_poisson_atlas_path
            return PoissonAtlas(self._poisson_atlas_path)

    @poisson_atlas.setter
    def poisson_atlas(self, value: str):
        self._poisson_atlas_path = value

    def add_sources_to_poisson(
        self, points: int, radii: units.Quantity, parallel_kwargs: dict = None
    ):
        """
        Add randomly sampled sources to the Poisson-atlas of this database instance.

        Parameters
        ----------
        points: int
            The number of randomly sampled points to add to the :py:class:`structures.map.PoissonAtlas` instance's ``COUNTS`` table.
        radii: :py:class:`astropy.units.Quantity`
            The search radii for each of the points. Can be either an ``(n,1)`` array where ``n`` is the number of points, or a scalar value.
        parallel_kwargs: dict
            Parameters for parallelization. For more information see :ref:`parallelization`.

            .. note::

                This method implements multi-threading in some subclasses.
        """
        mainlog.info(f"Generating random sample of {points} counts.")
        point_data = self.random_sample_count(
            points, radii, parallel_kwargs=parallel_kwargs
        )

        mainlog.info(f"Adding data to the Poisson map at {self._poisson_atlas_path}.")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.poisson_atlas.append_to_fits(point_data, "COUNTS")


class LocalDatabase(SourceDatabase):
    """
    Generic representation of a local database.
    """

    def __init__(self, table, db_name, **kwargs):
        self.table = table
        super().__init__(db_name, query_schema=self.table.schema, **kwargs)

    def __str__(self):
        return f"<LocalDatabase {self.name}>"

    def __repr__(self):
        return f"<LocalDatabase {self.name}, N={len(self.table)}>"

    def source_match_memory(
        self, source_table, search_radius=1 * units.arcmin, parallel_kwargs=None
    ):
        """
        Match a :py:class:`SourceTable` against this database.

        Parameters
        ----------
        source_table: :py:class:`SourceTable`
            The table to cross match against.
        search_radius: :py:class:`astropy.units.Quantity`
            The search radii for each of the source points.
        parallel_kwargs: dict
            Kwargs for parallelism.

            .. warning::

                This kwarg has no effect for local databases.

        Returns
        -------
        :py:class:`astropy.table.table.Table`
        """
        mainlog.info(
            f"Source matching table with {len(source_table)} entries against local database {self.name}."
        )

        # -- pull the coordinates from both tables -- #
        db_positions = self.table.get_coordinates()
        other_positions = source_table.get_coordinates()

        # ---------------------------------------------------#
        # Managing args and kwargs
        # ---------------------------------------------------#

        if not isinstance(search_radius, units.Quantity):
            mainlog.warning(
                f"Search radii is a data type without standard units ({type(search_radius)}). Defaulting to arcmin."
            )
            search_radius = search_radius * units.arcmin

        if not search_radius.isscalar:
            raise ValueError("The search radius for a local database must be a scalar.")

        _ = parallel_kwargs  # --> trick IDEs
        # ---------------------------------------------------#
        # Matching
        # ---------------------------------------------------#
        idxother, idxself, separation, _ = db_positions.search_around_sky(
            other_positions, search_radius
        )

        # ---------------------------------------------------#
        # Constructing the Table
        # ---------------------------------------------------#
        # Pull the base data provided by the local database.
        matched_table = self.table[idxself]  # --> this is the matched data.

        # Merge catalog and local database information.
        matched_table["CATOBJ"] = source_table[source_table.schema.NAME][idxother]
        matched_table["CATRA"], matched_table["CATDEC"] = (
            other_positions[idxother].ra.deg,
            other_positions[idxother].dec.deg,
        )
        matched_table["CATNMATCH"] = [
            len(list(set(idxself[idxother == idxo]))) for idxo in idxother
        ]

        return matched_table

    def source_match(
        self, path, source_table, search_radius=1 * units.arcmin, parallel_kwargs=None
    ):
        """
        Match a :py:class:`SourceTable` against this database.

        Parameters
        ----------
        path: str
            The path at which to write the match data.
        source_table: :py:class:`SourceTable`
            The table to cross match against.
        search_radius: :py:class:`astropy.units.Quantity`
            The search radii for each of the source points.
        parallel_kwargs: dict
            Kwargs for parallelism.

            .. warning::

                This kwarg has no effect for local databases.

        Returns
        -------
        None
        """
        import sqlalchemy as sql

        engine = sql.create_engine(f"sqlite:///{path}")
        matched_data = self.source_match_memory(
            source_table, search_radius=search_radius, parallel_kwargs=parallel_kwargs
        )
        matched_data.append_to_sql(f"{self.name}_MATCH", engine)

    def _query_radius(self, position, radius):
        # -- Pull the matches -- #
        positions = self.table.get_coordinates()
        return self.table[positions.separation(position) < radius]


class RemoteDatabase(SourceDatabase, ABC):
    """
    Generic class representation of a source database.
    """

    def __init__(self, db_name, **kwargs):
        super().__init__(db_name, **kwargs)

    def __str__(self):
        return f"<RemoteDatabase {self.name}>"

    def source_match(
        self, path, source_table, search_radii=1 * units.arcmin, parallel_kwargs=None
    ):
        """
        Match a :py:class:`SourceTable` against this database.

        Parameters
        ----------
        path: str
            The path at which to write the match data.
        source_table: :py:class:`SourceTable`
            The table to cross match against.
        search_radii: :py:class:`astropy.units.Quantity`
            The search radii for each of the source points.
        parallel_kwargs: dict
            Kwargs for parallelism.

        Returns
        -------
        None
        """
        import sqlalchemy as sql

        from pyXMIP.utilities.optimize import map_to_threads

        mainlog.info(f"Source matching {len(source_table)} against {self.name}.")

        # ---------------------------------------------------#
        # Managing args and kwargs
        # ---------------------------------------------------#
        engine = sql.create_engine(f"sqlite:///{path}")

        if not isinstance(search_radii, units.Quantity):
            mainlog.warning(
                f"Search radii is a data type without standard units ({type(search_radii)}). Defaulting to arcmin."
            )
            search_radii = np.array(search_radii) * units.arcmin

        if search_radii.isscalar:
            search_radii = search_radii * np.ones(len(source_table))

        # ---------------------------------------------------#
        # Running queries
        # ---------------------------------------------------#
        positions = source_table.get_coordinates()

        with logging_redirect_tqdm(loggers=[mainlog]):
            pbar = tqdm(desc=f"Querying {self.name}", total=len(search_radii))
            # -- Run once without pass to threads -- #
            result = map_to_threads(
                self._thread_pooled_source_match,
                positions,
                source_table,
                repeat(source_table.schema),
                search_radii,
                repeat(pbar),
                repeat(engine),
                threading_kw=parallel_kwargs,
            )

            for _ in result:
                pass

    def _thread_pooled_source_match(
        self,
        position,
        source_table_entry,
        source_table_schema,
        search_radius,
        pbar,
        engine,
    ):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                query = self.correct_query_output(
                    self._query_radius(position, search_radius)
                )
                if not len(query):
                    pbar.update()
                    return None

                query["CATOBJ"] = source_table_entry[source_table_schema.NAME]
                query["CATRA"] = position.ra.deg
                query["CATDEC"] = position.dec.deg
                query["CATNMATCH"] = len(query)
                query.meta = {}
                pbar.update()
            except Exception as exception:
                mainlog.error(exception.__repr__())
                return None

            with self._thread_lock:
                if not sql.inspect(engine).has_table(f"{self.name}_MATCH"):
                    mainlog.info(
                        f"[{threading.current_thread().name}] Creating table {self.name}_MATCH schema."
                    )
                    metadata = sql.MetaData()
                    _ = sql.Table(
                        f"{self.name}_MATCH",
                        metadata,
                        *[
                            sql.Column(k, convert_np_type_to_sql(v))
                            for k, v in dict(query.to_pandas().dtypes).items()
                        ],
                    )
                    metadata.create_all(engine)

                query.append_to_sql(f"{self.name}_MATCH", engine, method="multi")

        return None


class NED(RemoteDatabase):
    """
    Built-in remote database access to the `IPAC / CALTECH NED <https://ned.ipac.caltech.edu/>`_ database.

    Notes
    -----

    *from NED website:*

    NED is a comprehensive database of multiwavelength data for extragalactic objects, providing a systematic, ongoing
    fusion of information integrated from hundreds of large sky surveys and tens of thousands of research publications.
    The contents and services span the entire observed spectrum from gamma rays through radio frequencies.
    As new observations are published, they are cross-identified or statistically associated with previous data and
    integrated into a unified database to simplify queries and retrieval.
    Seamless connectivity is also provided to data in NASA astrophysics mission archives (IRSA, HEASARC, MAST),
    to the astrophysics literature via ADS, and to other data centers around the world.

    """

    default_poisson_atlas_path = os.path.join(poisson_map_directory, "NED.poisson.fits")
    default_query_config = {}
    default_query_schema = DEFAULT_SOURCE_SCHEMA_REGISTRY["NED"]

    # -- NED Settings -- #
    TIMEOUT = _DatabaseConfigSetting(default=60)
    """
    float: The maximum query response time to allow without raising an error.

    Default is 60 seconds.
    """

    def __init__(self, name="NED_STD", **kwargs):
        super().__init__(name, **kwargs)

        self.config_ned()

    def config_ned(self):
        Ned.TIMEOUT = self.TIMEOUT

    @classmethod
    def _default_correct_query_output(cls, table, schema=None):
        for col in table.columns:
            table[col].format = None
            if table[col].unit == "degrees":
                table[col].unit = "deg"

        table = super()._default_correct_query_output(table, schema=schema)

        return table

    def _query_radius(self, position, radius):
        """
        Query the remote database at the specified position and pull all sources within a given radius.

        Parameters
        ----------
        position: :py:class:`astropy.coordinates.SkyCoord`
            The position at which to query.
        radius: :py:class:`astropy.units.Quantity`
            The angular area about which to query.

        Returns
        -------
        :py:class:`astropy.table.Table`
        """
        # -- Attempt the query -- #
        try:
            output = SourceTable(Ned.query_region(position, radius))
        except requests.exceptions.ConnectionError:
            raise DatabaseError(
                f"Failed to complete query [{position},{radius}] to NED due to timeout."
            )

        # -- return data if valid -- #
        output.schema = self.query_schema
        return output

    def query_object(self, object_name):
        """
        Query NED for data related to a particular object.

        Parameters
        ----------
        object_name: str
            The name of the object to pull data for.

        Returns
        -------
        :py:class:`structures.table.SourceTable`
        """
        try:
            output = SourceTable(Ned.query_object(object_name))
        except requests.exceptions.ConnectionError:
            raise DatabaseError(
                f"Failed to complete query [{object_name}] to NED due to timeout."
            )

        # -- return data if valid -- #
        output.schema = self.query_schema
        return output

    @staticmethod
    def get_image_list(object_name, item="image"):
        """
        Query NED image or spectra urls.

        Parameters
        ----------
        object_name: str
            The name of the object to pull data for.

        Returns
        -------
        list of str
            The resulting URLs available for the object.
        """
        return Ned.get_image_list(object_name, item=item)

    @staticmethod
    def get_table(object_name, table):
        """
        Query NED for a particular data table related to an object.

        Parameters
        ----------
        object_name: str
            The name of the object to pull data for.
        table: str
            The table to query for.

        Returns
        -------
        :py:class:`astropy.table.table.Table`
            The resulting output table.
        """
        return Ned.get_table(object_name, table=table)


class SIMBAD(RemoteDatabase):
    """
    Built-in remote database access to the `SIMBAD <https://simbad.cds.unistra.fr/simbad/>`_ database.

    Notes
    -----

    *from SIMBAD website:*

    The SIMBAD database is managed by the Centre de Données astronomiques de Strasbourg (CDS).

    SIMBAD is the acronym for: **S** et of **I** dentifications, **M** easurments and **B** ibliography for
    **A** stronomical **D** ata.

    The SIMBAD software is developed by Marc Wenger (retired), Anaïs Oberto, Grégory Mantelet (CDS, Strasbourg).
    with contributions of students during trainings.

    The updating of SIMBAD is a continuous process.

    """

    default_poisson_atlas_path = os.path.join(
        poisson_map_directory, "SIMBAD.poisson.fits"
    )
    default_query_schema = DEFAULT_SOURCE_SCHEMA_REGISTRY["SIMBAD"]
    default_query_config = {}
    TIMEOUT = _DatabaseConfigSetting(default=120)
    """
    float: The maximum query response time to allow without raising an error.

    Default is 120 seconds.
    """
    EXTRA_COLUMNS = _DatabaseConfigSetting(default=["otypes", "ra(d;A)", "dec(d;D)"])
    """
    list: Additional VOTable columns to add to the returned query tables.

    Default is ``['otypes','ra(d;A)','dec(d;D)']``.
    """
    REMOVED_COLUMNS = _DatabaseConfigSetting(default=[])
    """
    list: VOTable columns to remove from the default list.

    Default is ``[]``.
    """
    ROW_LIMIT = _DatabaseConfigSetting(default=0)
    """
    int: The maximum allowed rows to return in a query.

    Default is ``0``, which allows for all results.
    """

    def __init__(self, name="SIMBAD_STD", **kwargs):
        super().__init__(name, **kwargs)

        self.config_simbad()

    @property
    def available_tables(self):
        """
        list: The tables accessible via SIMBAD.
        """
        return Simbad.list_tables()

    @staticmethod
    def clear_cache():
        """Clear the cache associated with SIMBAD queries."""
        Simbad.clear_cache()

    @staticmethod
    def list_columns(*args, **kwargs):
        """
        List the available columns.
        """
        return Simbad.list_columns(*args, **kwargs)

    @staticmethod
    def list_linked_tables(*args, **kwargs):
        """List the linked tables."""
        return Simbad.list_linked_tables(*args, **kwargs)

    def config_simbad(self):
        Simbad.TIMEOUT = self.TIMEOUT
        Simbad.ROW_LIMIT = self.ROW_LIMIT
        Simbad.add_votable_fields(*self.EXTRA_COLUMNS)
        Simbad.remove_votable_fields(*self.REMOVED_COLUMNS)

    @classmethod
    def _default_correct_query_output(cls, table: SourceTable, schema=None):
        if "RA" in table.columns:
            table["RA"] = Angle(table["RA"], unit="h").deg

        if "DEC" in table.columns:
            table["DEC"] = Angle(table["DEC"], unit="deg").deg

        table = super()._default_correct_query_output(table, schema=schema)

        return table

    def _query_radius(self, position: SkyCoord, radius: units.Quantity) -> SourceTable:
        """
        Query the remote database at the specified position and pull all sources within a given radius.
        Parameters
        ----------
        position: :py:class:`astropy.coordinates.SkyCoord`
            The position at which to query.
        radius: :py:class:`astropy.units.Quantity`
            The angular area about which to query.
        Returns
        -------
        :py:class:`astropy.table.Table`
        """
        # -- Attempt the query -- #
        try:
            output = SourceTable(Simbad.query_region(position, radius))
        except requests.exceptions.ConnectionError:
            raise DatabaseError(
                f"Failed to complete query [{position},{radius}] to Simbad due to timeout."
            )

        # -- return data if valid -- #
        output.schema = self.query_schema
        return output

    def query_object(self, object_name: str):
        """
        Query SIMBAD for data related to a particular object.

        Parameters
        ----------
        object_name: str
            The name of the object to pull data for.

        Returns
        -------
        :py:class:`structures.table.SourceTable`
        """
        try:
            output = SourceTable(Simbad.query_object(object_name))
        except requests.exceptions.ConnectionError:
            raise DatabaseError(
                f"Failed to complete query [{object_name}] to Simbad due to timeout."
            )

        # -- return data if valid -- #
        output.schema = self.query_schema
        return output

    @staticmethod
    def query_tap(query: str):
        """
        Query Simbad via ASQL query.

        Parameters
        ----------
        query: str
            The ASQL query to run.

        Returns
        -------
        :py:class:`astropy.table.table.Table`
            The resulting output table.
        """
        return Simbad.query_tap(query)


class DBRegistry(Registry):
    """
    A :py:class:`DBRegistry` instance is a collection of identifiable database classes.

    Notes
    -----

    The primary purpose of database registries is to allow the user and the backend to keep a "list" of the available
    databases.
    """

    def __init__(self, databases: list[SourceDatabase]):
        """
        Initialize a database registry.

        Parameters
        ----------
        databases: list of SourceDatabase
            The list of databases to load into the registry.

        """
        super().__init__({database.name: database for database in databases})

        # -- check that the classes are actually valid -- #
        assert all(isinstance(v, SourceDatabase) for v in self.values())

    @property
    def names(self) -> list[str]:
        return list(self.keys())

    @property
    def locals(self) -> list[str]:
        return [k for k, v in self.items() if isinstance(v, LocalDatabase)]

    @property
    def remotes(self) -> list[str]:
        return [k for k, v in self.items() if isinstance(v, RemoteDatabase)]

    @classmethod
    def _default_registry(cls):
        return cls([NED(), SIMBAD()])


DEFAULT_DATABASE_REGISTRY = DBRegistry._default_registry()

# ================================================================ #
# Additional Utility Functions                                     #
# ================================================================ #


def builtin_databases():
    """List the available built-in databases."""
    return DEFAULT_DATABASE_REGISTRY.names


def get_database(database_name, registry=None):
    """
    Fetch the initialized database from a given registry.
    Parameters
    ----------
    database_name: str
        The database name to fetch.
    registry: DBRegistry, optional
        The registry to fetch from. If not specified, will search through the default registry.

    Returns
    -------
    SourceDatabase
        The output database.
    """
    if not registry:
        return DEFAULT_DATABASE_REGISTRY.get(database_name)
    else:
        return registry.get(database_name)


def get_poisson_path(database_name, registry=None):
    """
    Fetch the path to the Poisson Atlas for a given database.
    Parameters
    ----------
    database_name: str
        The name of the database to fetch.
    registry: DBRegistry, optional
        The registry to search.

    Returns
    -------
    str
    """
    return get_database(database_name, registry=registry)._poisson_atlas_path


def get_poisson_atlas(database_name, registry=None):
    """
    Fetch the path to the Poisson Atlas for a given database.
    Parameters
    ----------
    database_name: str
        The name of the database to fetch.
    registry: DBRegistry, optional
        The registry to search.

    Returns
    -------
    str
    """
    return get_database(database_name, registry=registry).poisson_atlas


def add_points_to_poisson_map(database_name, n, radii, registry=None, **kwargs):
    """
    Add new points to an existing :py:class:`structures.map.PoissonAtlas` attached to a database.

    Parameters
    ----------
    database_name: str
        The database to load.
    n: int
        The number of new points to add.
    radii: Quantity
        Value (scalar or vector) for the radius of search for each sample.
    registry: DBRegistry, optional
        The registry to search. If not provided, searches the default registry.
    **kwargs:
        Additional kwargs to pass to the random counting process.

    Returns
    -------
    None
    """
    if isinstance(radii, str):
        radii = float(radii.split(" ")[0]) * units.Unit(radii.split(" ")[1])
    elif isinstance(radii, (float, int)):
        radii = radii * units.arcmin

    database = get_database(database_name, registry)

    database.add_sources_to_poisson(n, radii, **kwargs)



"""
Testing utilities for pyXMIP
"""
import os
import pathlib as pt

from astropy.table import Table


def check_astropy_table(table, answer_dir, subpath, answer_store):
    path = pt.Path(os.path.join(answer_dir, subpath))

    if not path.parents[0].exists():
        path.parents[0].mkdir(parents=True)

    if answer_store or not path.exists():
        table.write(path, format="ascii", overwrite=True)
        return None

    else:
        old_table = Table.read(path, format="ascii")

        assert (
            old_table.to_pandas() == table.to_pandas()
        ).all, f"There was a change in the table at {path}."



"""
Plotting utilities for use in the pyXMIP backend.
"""
import functools

import healpy as hp
import healpy.projaxes as PA
import numpy as np
from matplotlib import pyplot as plt

from pyXMIP.utilities.core import pxconfig


# ======================================================================================================================#
# STYLE FUNCTIONS                                                                                                      #
# ======================================================================================================================#
def _enforce_style(func):
    """Enforces the mpl style."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        _rcp_copy = plt.rcParams.copy()

        for _k, _v in pxconfig["plotting"]["defaults"].items():
            plt.rcParams[_k] = _v

        out = func(*args, **kwargs)

        plt.rcParams = _rcp_copy
        del _rcp_copy

        return out

    return wrapper


def set_style():
    """
    Uses the ``pyXMIP`` settings in the configuration file and enforces those formatting choices on matplotlib.
    """
    for _k, _v in pxconfig["plotting"]["defaults"].items():
        plt.rcParams[_k] = _v


# ===================================================================================================================== #
# Minor Process Functions                                                                                               #
# ===================================================================================================================== #
_projection_classes = {"mollweide": PA.HpxMollweideAxes}


def _check_healpix(map):
    hp.pixelfunc.check_nside(hp.pixelfunc.get_nside(map))

    return hp.pixelfunc.get_nside(map)


def get_hips_image(
    center, FOV, dims, hips_path="CDS/P/DSS2/red", projection="AIT", **kwargs
):
    """
    Fetch a HIPs image from arbitrary servers and return.

    Parameters
    ----------
    center: :py:class:`astropy.coordinates.SkyCoord`
        The coordinates corresponding to the center of the image.
    FOV: :py:class:`astropy.coordinates.Angle` or str or float
        The field of view for the image. This will be automatically converted the :py:class:`astropy.coordinates.Angle` regardless
        of the input type. If conversion fails to occur, an error is raised.
    dims: tuple of int
        The dimensions of the returned image.
    hips_path: str, optional
        The server path to the HIPs image desired.
    projection: str, optional
        The default return projection for the output image.

    """
    from astropy.coordinates import Angle, Latitude, Longitude
    from astroquery.hips2fits import hips2fits

    # ============================ #
    # Setup / build WCS
    # ============================ #
    parameters = {
        "hips": hips_path,
        "ra": Longitude(center.transform_to("icrs").ra),
        "dec": Latitude(center.transform_to("icrs").dec),
        "width": dims[0],
        "height": dims[1],
        "fov": Angle(FOV),
        "format": "fits",
        "projection": projection,
        **kwargs,
    }

    return hips2fits.query(**parameters)


# ======================================================================================================================#
# Plotting Systems                                                                                                      #
# ======================================================================================================================#


def plot_healpix(healpix_map, fig=None, ax=None, projection=None, **kwargs):
    # -------------------------------------------- #
    # Map checking and setup
    # -------------------------------------------- #
    # managing the projection type.
    if projection is None:
        projection = "mollweide"
    projection_class = _projection_classes[projection]

    # Setting up the figure and axes
    if fig is None:
        fig = plt.figure(figsize=kwargs.pop("figsize", (8, 4)))
    if ax is None:
        ax = projection_class(
            fig,
            [0.025, 0.025, 0.95, 0.95],
            coord=kwargs.pop("coord", None),
            rot=kwargs.pop("rot", None),
            format=kwargs.pop("format", "%g"),
            flipconv=kwargs.pop("flip", "geo"),
        )
    elif not isinstance(ax, projection_class):
        # ax is a class, but not the right one.
        _old_axes_position = ax._position
        fig.remove(ax)
        del ax

        # create the new axes.
        ax = projection_class(
            fig,
            _old_axes_position,
            coord=kwargs.pop("coord", None),
            rot=kwargs.pop("rot", None),
            format=kwargs.pop("format", "%g"),
            flipconv=kwargs.pop("flip", "geo"),
        )
    else:
        pass

    fig.add_axes(ax)
    # -------------------------------------------- #
    # Adding the map to the axes.
    # -------------------------------------------- #
    img = ax.projmap(
        healpix_map,
        vmin=kwargs.pop("vmin", np.amin(healpix_map)),
        vmax=kwargs.pop("vmax", np.amax(healpix_map)),
        cmap=kwargs.pop("cmap", "viridis"),
        badcolor=kwargs.pop("fillna", "black"),
        bgcolor=kwargs.pop("facecolor", "w"),
        norm=kwargs.pop("norm", None),
        alpha=kwargs.pop("alpha", None),
    )

    return fig, ax, img


def plot_hips(
    center,
    FOV,
    fig=None,
    projection=None,
    resolution=1000,
    hips_path="CDS/P/DSS2/red",
    hips_kwargs=None,
    **kwargs
):
    # -- Managing WCS axes -- #
    from astropy.wcs import WCS

    if projection is None:
        projection = "AIT"

    # ========================================== #
    # Pull Image                                 #
    # ========================================== #
    if hips_kwargs is None:
        hips_kwargs = {}

    image = get_hips_image(
        center,
        FOV,
        (resolution, resolution),
        hips_path=hips_path,
        projection=projection,
        **hips_kwargs
    )

    # -- Throw the WCS coordinates -- #
    wcs = WCS(image[0].header)

    # ========================================== #
    # Setup the figure and axes as needed        #
    # ========================================== #
    if fig is None:
        fig = plt.figure(figsize=kwargs.pop("figsize", (6, 6)))

    ax = fig.add_subplot(111, projection=wcs)
    ax.imshow(image[0].data, **kwargs)

    return fig, ax, wcs



r"""
Module for representing functions on the sky.

Notes
-----

The :py:mod:`structures.map` module provides an interface for managing data on the sky. The :py:class:`structures.map.MapAtlas` class
represents a collection of :py:class:`structures.map.Map` objects. Each of these maps acts like a function of sky position.
"""
import pathlib as pt
import warnings
from time import asctime
from types import SimpleNamespace

import healpy as hp
import numpy as np
from _collections_abc import Collection
from astropy import coordinates as astro_coords
from astropy import units
from astropy.io import fits
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from pyXMIP.utilities.core import enforce_units
from pyXMIP.utilities.geo import convert_coordinates, convert_skycoord
from pyXMIP.utilities.logging import mainlog
from pyXMIP.utilities.plot import _enforce_style, plot_healpix


class _AtlasHeaderParam:
    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        try:
            val = getattr(instance, f"_{self._name}")
            if val is not None:
                return val
            else:
                raise AttributeError
        except AttributeError:
            with fits.open(instance.path) as hudl:
                setattr(instance, f"_{self._name}", hudl[0].header[self._name])
                return hudl[0].header[self._name]

    def __set__(self, instance, value):
        with fits.open(instance.path, "update") as hudl:
            hudl[0].header[self._name] = value
            hudl.flush()


class _MapHeaderParam:
    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        try:
            val = getattr(instance, f"_{self._name}")
            if val is not None:
                return val
            else:
                raise AttributeError
        except AttributeError:
            with fits.open(instance.path) as hudl:
                setattr(
                    instance, f"_{self._name}", hudl[instance.name].header[self._name]
                )
                return hudl[instance.name].header[self._name]

    def __set__(self, instance, value):
        with fits.open(instance.path, "update") as hudl:
            hudl[instance.name].header[self._name] = value
            hudl.flush()


class MapAtlas:
    r"""
    The :py:class:`MapAtlas` class is a generic wrapper for ``.fits`` files which is designed to store sky maps and associated
    data tables.

    Notes
    -----

    On it's surface, :py:class:`MapAtlas` is a standard fits file; except for a few special details regarding the headers and the
    data formats. In the primary HDU, the fits file has additional parameters which determine the behavior of the underlying sky map geometry.

    - ``NPIX``: The number of sky-pixels in the underlying HEALPix grid.
    - ``NSIDE``: Equivalent to ``NPIX`` - specifies the HEALPix grid.
    - ``CSYS``: The native coordinate system of the map.
    - ``CDATE``: The date the :py:class:`MapAtlas` object was created.
    - ``EDATE``: The last date the :py:class:`MapAtlas` was edited.
    - ``RES``: The resolution of the map.

    .. note::

        HEALPix maps use a different coordinate convention than is standard for spherical coordinate systems. Generically, HEALPix uses
        coordinates :math:`0\le \phi < 2\pi` and :math:`0 \le \theta < \pi`, with :math:`\theta` measured from the north-pole.

        The user **NEVER** interacts with this coordinate system when interacting with package objects. The coordinates are automatically converted
        to the standard convention for lon / lat: :math:`[0,2\pi] \times [-\pi/2,\pi/2]`. These are the so-called ``base_coordinates`` of the
        HEALPix grid and are not yet tied to any physical context. The ``CSYS`` header entry specifies the :py:mod:`astropy.coordinates` frame
        on which those coordinates are to be interpreted.

    In addition to the Primary HDU, which has the aforementioned special headers, :py:class:`MapAtlas` instances can contain
    very special Image HDU's called Map HDU's. These are ``1xNPIX`` arrays representing some function on the HEALPix grid.

    The headers of Map HDU's need to be identified in their header with ``ISMAP = True``.
    """
    NPIX = _AtlasHeaderParam()
    NSIDE = _AtlasHeaderParam()
    CSYS = _AtlasHeaderParam()
    CD = _AtlasHeaderParam()
    ED = _AtlasHeaderParam()

    def __init__(self, filepath):
        """
        Initialize the :py:class:`MapAtlas` from a specified ``.fits`` file.

        Parameters
        ----------
        filepath: str
            The path to the ``.fits`` file to read from.

        """
        self.path = filepath

        if not pt.Path(self.path).exists():
            raise FileNotFoundError(f"There is no Atlas file at {self.path}.")

    def get_map(self, name):
        """
        Obtain an instance of the map corresponding the the name specified from this Atlas.

        Parameters
        ----------
        name: str
            The name of the map.

        Returns
        -------
        :py:class:`Map`

        Examples
        --------

        Let's try loading the galaxies poisson map for the SIMBAD database.

        .. code-block:: python

            from pyXMIP.structures.databases import SIMBAD
            poisson_atlas = SIMBAD.get_default_poisson_atlas() # load the default poisson map.
            map = poisson_atlas.get_map('G')

        """
        return Map(self.path, name.upper())

    @property
    def map_names(self):
        """
        Returns a list of the available map HDU names.

        Returns
        -------
        list
        """
        with fits.open(self.path, "update") as hudl:
            return [
                q.name
                for q in hudl
                if isinstance(q, fits.ImageHDU) and q.header["ISMAP"]
            ]

    @property
    def has_maps(self):
        return True if len(self.map_names) else False

    @property
    def hdus(self):
        """
        Return a list of all of the available HDUs.

        Returns
        -------
        list
        """
        with fits.open(self.path, "update") as hudl:
            return [u.name for u in hudl]

    @property
    def coordinate_frame(self):
        """The :py:class:`astropy.coordinates.GenericFrame` for the Atlas's coordinate system"""
        return getattr(astro_coords, self.CSYS)

    @property
    def pixel_positions(self):
        """The SkyCoord positions of the healpix pixels."""
        _th, _ph = hp.pix2ang(self.NSIDE, np.arange(self.NPIX))
        _ph, _th = convert_coordinates(
            _ph, _th, from_system="healpix", to_system="latlon"
        )
        return astro_coords.SkyCoord(_ph, _th, frame=self.coordinate_frame, unit="rad")

    @classmethod
    def generate(cls, path, resolution, overwrite=False):
        """
        Create an empty :py:class:`SkyAtlas` of a given resolution.

        Parameters
        ----------
        path: str
            The path to the ``.fits`` file from which this object is to be loaded / saved.
        resolution: :py:class:`astropy.units.Quantity` or Number
            The resolution of the HEALPix grid. If a value is passed with units, the units are assumed to be in ``rad``.
        overwrite: bool
            Allow file overwrite.

        Returns
        -------
        :py:class:`SkyAtlas`
        """
        from astropy.io import fits

        resolution = enforce_units(resolution, units.rad)

        mainlog.info(
            f"Generating blank SkyAtlas with resolution {resolution} at {path}."
        )

        # -- resolving the grid -- #
        n_sides = int(np.ceil(1 / (resolution.to_value(units.rad) * np.sqrt(3))))
        n_pixels = hp.nside2npix(n_sides)

        # -- generating the meta data -- #
        header = fits.Header()

        header["CDATE"] = asctime()
        header["EDATE"] = asctime()
        header["NSIDE"] = n_sides
        header["NPIX"] = n_pixels
        header["RES"] = resolution.to_value(units.rad)
        header["CSYS"] = "ICRS"

        # -- creating the fits file -- #
        empty_primary = fits.PrimaryHDU(header=header)
        hudl = fits.HDUList([empty_primary])
        hudl.writeto(path, overwrite=overwrite)

        return cls(path)

    def reshape_healpix(self, resolution, force=False):
        """
        Reshape the underlying HEALPix grid for the data.

        .. warning::

            This will result in the deletion of the maps in the Atlas.

        Parameters
        ----------
        resolution: units.Quantity
            The resolution radius of the new HEALPix grid.
        force: bool, optional
            If ``True``, force the deletion of maps instead of failing when they are encountered.

        Returns
        -------
        None
        """
        mainlog.info(
            f"Reshaping HEALPix grid to resolution {resolution} [{self.path}]."
        )

        # ----------------------------------------------------------#
        # Managing existing HEALPix grids that need to be replaced
        # ----------------------------------------------------------#
        if not force and self.has_maps:
            raise ValueError(
                "Maps already exist in this atlas. They will be removed if you proceed. To proceed use force=True."
            )
        elif self.has_maps:
            mainlog.warning(
                f"Deleted {len(self.map_names)} maps from {self.path} to change HEALPix size."
            )
            with fits.open(self.path, "update") as hudl:
                for name in self.map_names:
                    del hudl[name]

                hudl.flush()
        else:
            pass
        # ----------------------------------------------------------#
        # Changing the HEALPix geometry.
        # ----------------------------------------------------------#
        n_sides = int(np.ceil(1 / (resolution.to_value(units.rad) * np.sqrt(3))))
        n_pixels = hp.nside2npix(n_sides)

        with fits.open(self.path, "update") as hudl:
            hudl[0].header["NSIDE"] = n_sides
            hudl[0].header["NPIX"] = n_pixels
            hudl[0].header["RES"] = resolution.to_value(units.rad)
            hudl[0].header["EDATE"] = asctime()

            hudl.flush()

    def remove(self):
        mainlog.info(f"Removing {self.path}.")
        pt.Path(self.path).unlink()
        del self

    @classmethod
    def _get_descriptors(cls):
        # fetching descriptors.
        return [m for m, v in cls.__dict__.items() if isinstance(v, _AtlasHeaderParam)]

    def _update_attributes(self):
        # fetch the attached descriptors
        descriptors = self._get_descriptors()

        for descriptor in descriptors:
            setattr(self, f"_{descriptor._name}", None)  # resets everything.


class StatAtlas(MapAtlas):
    """
    The :py:class:`StatAtlas` object is a subclass of :py:class:`MapAtlas` which wraps a ``.fits`` file containing (in addition to the
    standard maps), a table of random samples from a specified database.

    Notes
    -----

    On it's surface, :py:class:`MapAtlas` is a standard fits file; except for a few special details regarding the headers and the
    data formats. In the primary HDU, the fits file has additional parameters which determine the behavior of the underlying sky map geometry.

    - ``NPIX``: The number of sky-pixels in the underlying HEALPix grid.
    - ``NSIDE``: Equivalent to ``NPIX`` - specifies the HEALPix grid.
    - ``CSYS``: The native coordinate system of the map.
    - ``CDATE``: The date the :py:class:`MapAtlas` object was created.
    - ``EDATE``: The last date the :py:class:`MapAtlas` was edited.
    - ``RES``: The resolution of the map.
    - ``DBNAME``: The name of the database.
    """

    DBNAME = _AtlasHeaderParam()

    def __init__(self, filepath):
        super().__init__(filepath)

    def __len__(self):
        return len(self.COUNTS)

    @property
    def COUNTS(self):
        from astropy.table import Table

        with warnings.catch_warnings(), fits.open(self.path) as hudl:
            warnings.simplefilter("ignore")
            return Table(hudl["COUNTS"].data)

    def get_points(self):
        """
        Fetch the underlying ``COUNTS`` table for the atlas.

        Returns
        -------
        :py:class:`astropy.table.Table`
        """
        import warnings

        from astropy.table import Table

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with fits.open(self.path) as hudl:
                _out = Table(hudl["COUNTS"].data)

        # determining the HEALPix grid
        # !We ALWAYS write counts in RA/DEC for simplicity.
        count_positions = astro_coords.SkyCoord(
            ra=_out["RA"], dec=_out["DEC"], unit="deg"
        )

        _p, _t = (
            count_positions.transform_to(self.coordinate_frame).spherical.lon.rad,
            count_positions.transform_to(self.coordinate_frame).spherical.lat.rad,
        )
        _p, _t = convert_coordinates(_p, _t, from_system="latlon", to_system="healpix")

        _out["PIX_ID"] = hp.ang2pix(self.NSIDE, _t, _p)

        return _out

    def sample_from_database(self, npoints, search_radius, *args, **kwargs):
        """
        Randomly sample points from the database specified in the meta-data of this atlas and add the count
        data to the pre-existing COUNTS table.

        Parameters
        ----------
        npoints: int
            The number of random sample points to query for.
        search_radius: :py:class:`astropy.units.Quantity`
            Angular search radius for counting around each point.
        args
            Additional arguments to pass through.
        kwargs
            Additional key-word arguments to pass through.

        Returns
        -------
        None
        """
        from pyXMIP.structures.databases import DEFAULT_DATABASE_REGISTRY

        mainlog.info(f"Adding {npoints} counting queries to {self.path} atlas.")

        # -- fetching the correct database -- #
        _registry = kwargs.pop("registry", DEFAULT_DATABASE_REGISTRY)
        if self.DBNAME == "NONE":
            raise KeyError(
                "This PoissonAtlas doesn't appear to be linked to a database. It may be corrupted or it was never set. Use instance.DBNAME = 'new_name' to set a database name."
            )
        elif self.DBNAME not in _registry:
            raise KeyError(
                "This Database doesn't correspond to a known database in the registry provided."
            )
        else:
            database = _registry[self.DBNAME]
        mainlog.debug(f"LINKED DATABASE: {database}")

        # -- Requesting pull from database -- #
        sample_count_table = database.random_sample_count(
            npoints, search_radius, *args, **kwargs
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.append_to_fits(sample_count_table, "COUNTS")

        mainlog.info(f"Appended {npoints} samples to COUNTS.")

    def reset(self, resolution, **kwargs):
        mainlog.info(f"Resetting Atlas at {self.path}.")
        path = self.path
        kwargs["database"] = kwargs.get("database", self.DBNAME)
        self.remove()
        self.generate(path, resolution, **kwargs)

    def _write_build_output_to_fits(self, object_type, output_object, overwrite=False):
        # =========================================== #
        # Writing the map to an HDU of the correct name
        object_type = object_type.upper()
        with fits.open(self.path, "update") as hdul:
            if object_type in [hdu.name for hdu in hdul] and not overwrite:
                # The map already exists and we cannot overwrite.
                raise ValueError(
                    f"The map {object_type} already exists in {self.path} and overwrite = False."
                )
            elif object_type in [hdu.name for hdu in hdul]:
                # The map exists we need to remove it.
                del hdul[object_type]

            # -- Writing the new map -- #
            image_hdu = fits.ImageHDU(output_object.map)
            image_hdu.name = object_type
            image_hdu.header["ISMAP"] = True
            image_hdu.header["METH"] = output_object.method
            image_hdu.header["DATE"] = asctime()
            image_hdu.header["CSYS"] = self.CSYS

            hdul.append(image_hdu)
            hdul.flush()

    @classmethod
    def generate(cls, path, resolution, overwrite=False, database="NONE"):
        obj = super().generate(path, resolution, overwrite=overwrite)
        obj.DBNAME = database

        return obj

    def append_to_fits(self, table, hudl):
        _self_hudl = fits.table_to_hdu(table)

        with fits.open(self.path, "update") as hudl_list:
            if hudl in hudl_list:
                _hudl, _len_hudl = hudl_list[hudl], len(hudl_list[hudl].data)
                new_hudl = fits.BinTableHDU.from_columns(
                    _hudl.columns, nrows=_len_hudl + len(table)
                )
                for colname in hudl_list[hudl].columns.names:
                    new_hudl.data[colname][_len_hudl:] = _self_hudl.data[colname]

                del hudl_list[hudl]
            else:
                new_hudl = fits.table_to_hdu(table)

            new_hudl.name = hudl
            hudl_list.append(new_hudl)
            hudl_list.flush()


class PoissonAtlas(StatAtlas):
    """
    Special case class of :py:class:`StatAtlas` for Poisson maps.
    """

    def build_poisson_map(
        self, object_type, method="KNN", overwrite=True, inplace=True, *args, **kwargs
    ):
        r"""
        Build a Poisson map in the Atlas for a specific object type.

        Parameters
        ----------
        object_type: str
            A particular object type in this PoissonAtlas from which to generate the map.
        method: str
            The method by which to calculate the map.
        overwrite: bool, optional
            [Default ``True``]. Allow the overwriting of pre-existing maps of the same name?
        inplace: bool, optional
            [Default ``True``]. If ``True``, the map is written directly to file. If ``False`` it is returned to the user.
        *args
            Additional arguments.
        **kwargs
            Additional keyword arguments.

        Returns
        -------

        """
        # ======================================================= #
        # SETUP
        # ======================================================= #
        _method = getattr(self, f"build_poisson_map_{method}")
        mainlog.info(f"Generating {object_type} map using {method}.")

        # ======================================================= #
        # RUNNING
        # ======================================================= #
        _output = _method(object_type, *args, **kwargs)

        # ===================================================== #
        # Writing data
        # ===================================================== #
        if inplace:
            self._write_build_output_to_fits(object_type, _output, overwrite=overwrite)
        else:
            return _output

    def build_poisson_maps(
        self,
        object_types="all",
        methods="MAP",
        overwrite=True,
        build_args=None,
        build_kwargs=None,
        **kwargs,
    ):
        r"""
        Construct all or a subset of the available object type Poisson maps for this Atlas.

        Parameters
        ----------
        object_types: list of str or str, optional
            The object types to generate Poisson maps for. If ``all`` (default), then all of the object types in the ``COUNTS``
            table are used.
        methods: list of str or str, optional
            The method to use for the map generation. If this is a list, it must match the length of ``object_types`` and each
            of the Poisson maps will be generated using the specified method. If ``methods`` is a single ``str``, then a single
            method will be used for all of the generation procedures.
        build_args: list of list, optional
            List of additional arguments to pass to each of the generation methods individually. By default, this is ``None``, and will result
            in empty lists being appended to the arguments of each generation method. If specified, this parameter must be a list of length ``object_types.size``
            with each element also being a list (or arbitrary size) containing any additional args to pass.
        build_kwargs: list of dict, optional
            Similar to ``build_args``, ``build_kwargs`` may be specified to pass particular keyword arguments along to the individual generation methods.
            If left unspecified, no kwargs are passed to the sub-generation processes.
        overwrite: bool, optional
            [Default ``True``]. Allow the overwriting of pre-existing maps of the same name?
        **kwargs
            Additional keyword arguments.

        Returns
        -------

        """
        # ======================================================================== #
        # Setup
        # ======================================================================== #
        _count_data = self.get_points()  # fetch the point data to process.

        # -- managing args / kwargs -- #
        if object_types == "all":
            object_types = list(
                _count_data.columns[:-5]
            )  # Removing RA,DEC,RAD,PIX_ID,TIME
        else:
            object_types = [
                object_type
                for object_type in object_types
                if object_type in _count_data.columns
            ]

        mainlog.info(f"Generating Poisson maps for {len(object_types)} object types...")

        # -- Managing computation methods -- #
        if isinstance(methods, str):
            methods = [methods] * len(object_types)
        elif isinstance(methods, Collection):
            assert len(methods) == len(
                object_types
            ), f"Attempted to pass {len(methods)} methods for {len(object_types)} object types."
        else:
            raise TypeError(
                f"Kwarg `methods` has type {type(methods)} when it was expected to be `str` or `Collection`."
            )

        # Managing the build args / kwargs
        if build_args is None:
            build_args = [[] for _ in range(len(object_types))]
        else:
            assert isinstance(
                build_args, Collection
            ), f"Kwarg `build_args` has type {type(build_args)} which is not a subclass of `Collection`."

        if build_kwargs is None:
            build_kwargs = [{} for _ in range(len(object_types))]
        else:
            assert isinstance(
                build_args, Collection
            ), f"Kwarg `build_kwargs` has type {type(build_kwargs)} which is not a subclass of `Collection`."

        # ======================================================================== #
        # Map Generation Processes
        # ======================================================================== #
        with logging_redirect_tqdm(loggers=[mainlog]):
            for object_type, method, bargs, bkwargs in tqdm(
                zip(object_types, methods, build_args, build_kwargs),
                total=len(object_types),
                desc="Constructing Poisson Maps",
                disable=(not kwargs.get("progress_bar", True)),
                leave=True,
            ):
                bkwargs["progress_bar"] = kwargs.get(
                    "progress_bar", bkwargs.get("progress_bar", True)
                )

                self.build_poisson_map(
                    object_type,
                    method=method,
                    overwrite=overwrite,
                    inplace=True,
                    *bargs,
                    **bkwargs,
                )

    def build_poisson_map_regressor(
        self,
        regressor,
        object_type,
        retrain=True,
        cross_validate=True,
        training_kw=None,
        param_kw=None,
        cv_kw=None,
        **kwargs,
    ):
        """
        Build a poisson map with a regressor.
        Parameters
        ----------
        regressor
        retrain
        cross_validate
        training_kw
        param_kw

        Returns
        -------

        """
        # =================================================== #
        # SETUP
        # =================================================== #
        phi, theta = convert_skycoord(
            self.pixel_positions.transform_to("icrs"), "math_convention"
        )
        map_positions = np.vstack(
            [theta, phi]
        ).T  # --> This is done in oposite order for haversine.

        # get the points that are being fitted.
        count_data = self.get_points()

        if cv_kw is None:
            cv_kw = {}
        # ================================================= #
        # Cross Validation
        # ================================================= #
        if cross_validate:
            cross_validation_data = regressor.cross_validate(
                count_data,
                object_type,
                training_kw=training_kw,
                param_kw=param_kw,
                **cv_kw,
            )

            # Parse the data and find the best score
            test_scores = cross_validation_data["mean_test_score"]
            best_score, best_params = (
                np.amax(test_scores),
                cross_validation_data["params"][
                    list(cross_validation_data["rank_test_score"]).index(1)
                ],
            )

            if kwargs.get("cv_score_threshold", None) is not None:
                # the user has specified a cross validation score threshold.
                _threshold = kwargs.pop("cv_score_threshold")
                assert (
                    best_score >= _threshold
                ), f"Score threshold was {_threshold} and best CV score was {best_score}."

            mainlog.debug(f"CROSS-VALIDATED {object_type}. PARAMS:")
            for k, v in best_params.items():
                mainlog.debug(f"{k} = {v}")

            regressor.regressor.set_params(**best_params)
        else:
            if param_kw is not None:
                regressor.regressor.set_params(**{k: v[0] for k, v in param_kw.items()})
        # ================================================= #
        # Training
        # ================================================= #
        if retrain:
            training_kw = {} if training_kw is None else training_kw
            score = regressor.train_model(count_data, object_type, **training_kw)

            if kwargs.get("score_threshold", None) is not None:
                # the user has specified a cross validation score threshold.
                _threshold = kwargs.pop("score_threshold")
                assert (
                    score >= _threshold
                ), f"Score threshold was {_threshold} and best test score was {score}."

        # ================================================= #
        # Output
        # ================================================= #
        _out_map = regressor.regressor.predict(map_positions)

        return SimpleNamespace(map=_out_map, method=regressor.__class__.__name__)

    def build_poisson_map_KNN(self, object_type, *args, **kwargs):
        from pyXMIP.stats.map_regression import KNNeighborMapRegressor

        _ = args

        return self.build_poisson_map_regressor(
            KNNeighborMapRegressor(), object_type, **kwargs
        )

    def build_poisson_map_RNN(self, object_type, *args, **kwargs):
        from pyXMIP.stats.map_regression import RNNeighborMapRegressor

        _ = args

        return self.build_poisson_map_regressor(
            RNNeighborMapRegressor(), object_type, **kwargs
        )

    def build_poisson_map_MAP(self, object_type, *args, **kwargs):
        from pyXMIP.stats.map_regression import BayesianPoissonMapRegressor

        _ = args
        # -- pull counts data -- #
        count_table = self.get_points()

        # -- setup the regressor -- #
        regressor = BayesianPoissonMapRegressor(
            prior=kwargs.pop("prior", None), model=kwargs.pop("model", None)
        )

        estimates = regressor.build_map_MAP(count_table, object_type, **kwargs)

        # -- returning the map -- #
        map = np.zeros(self.NPIX)
        map[: len(estimates)] = estimates

        return SimpleNamespace(map=map, method=regressor.__class__.__name__)


class Map:
    """
    Representation of a callable HEALPix map in an Atlas.
    """

    NPIX = _AtlasHeaderParam()
    NSIDE = _AtlasHeaderParam()
    CSYS = _MapHeaderParam()
    CD = _MapHeaderParam()
    ED = _MapHeaderParam()

    def __init__(self, path, name):
        """
        Initialize the specified Map HDU from the atlas located at ``path``.

        Parameters
        ----------
        path: str
            The path to the :py:class:`MapAtlas` instance.
        name: str
            The name of the HDU to load.
        """
        mainlog.debug(f"Loading Map object from {path} [Name={name}].")

        # -- loading basic attributes -- #
        self.path = pt.Path(path)
        self.name = name.upper()

        # -- reading the values -- #
        with fits.open(self.path) as hudl:
            assert name in [
                u.name for u in hudl
            ], f"There is no SkyMap {name} in {path}."

            self._data = hudl[name].data

    def __call__(self, position):
        """
        Evaluate the skymap.

        Parameters
        ----------
        position: :py:class:`astropy.coordinates.SkyCoord`

        """
        pixels = self.get_healpix_id(position)

        return self.data[pixels]

    @classmethod
    def _get_descriptors(cls):
        # fetching descriptors.
        return [
            m
            for m, v in cls.__dict__.items()
            if isinstance(v, (_AtlasHeaderParam, _MapHeaderParam))
        ]

    def _update_attributes(self):
        # fetch the attached descriptors
        descriptors = self._get_descriptors()

        for descriptor in descriptors:
            setattr(self, f"_{descriptor}", None)  # resets everything.

    @property
    def data(self):
        try:
            val = self._data
            if val is not None:
                return val
            else:
                raise AttributeError
        except AttributeError:
            with fits.open(self.path) as hudl:
                self._data = hudl[self.name].data
                return self._data

    @data.setter
    def data(self, value):
        with fits.open(self.path, "update") as hudl:
            hudl[self.name].data = value
            hudl.flush()

        self.ED = asctime()

    @property
    def coordinate_frame(self):
        """The :py:class:`astropy.coordinates.GenericFrame` for the Atlas's coordinate system"""
        with fits.open(self.path) as hudl:
            return getattr(astro_coords, hudl[self.name].header["CSYS"])

    def transform_map_coordinates(self, frame, inplace=False):
        # -- coercing the frame type -- #
        if isinstance(frame, str):
            frame = getattr(astro_coords, frame)
        else:
            pass

        # -- converting -- #
        _th, _ph = hp.pix2ang(self.NSIDE, np.arange(self.NPIX))
        _ph, _th = convert_coordinates(
            _ph, _th, from_system="healpix", to_system="latlon"
        )
        coords = astro_coords.SkyCoord(_ph, _th, frame=frame, unit="rad")
        output_array = self(coords)

        if inplace:
            with fits.open(self.path, "update") as hudl:
                hudl[self.name].header["CSYS"] = frame.__name__
                hudl[self.name].data = output_array
                hudl.flush()
            self.ED = asctime()
            self._update_attributes()
            self._data = output_array

        return output_array

    @property
    def pixel_positions(self):
        """The SkyCoord positions of the healpix pixels."""
        _th, _ph = hp.pix2ang(self.NSIDE, np.arange(self.NPIX))
        _ph, _th = convert_coordinates(
            _ph, _th, from_system="healpix", to_system="latlon"
        )
        return astro_coords.SkyCoord(_ph, _th, frame=self.coordinate_frame, unit="rad")

    def get_healpix_id(self, positions, frame=None):
        if frame is None:
            frame = self.coordinate_frame

        positions = positions.transform_to(frame)
        _ph, _th = convert_skycoord(positions, "healpix")
        print(_ph, _th)
        ids = hp.ang2pix(self.NSIDE, _th, _ph)

        return ids

    @_enforce_style
    def plot(self, *args, **kwargs):
        return plot_healpix(self.data, *args, **kwargs)


def _parse_default_kwarg_groups(defaults, kwargs):
    # -- setting args and kwargs -- #
    out = {k: kwargs.pop(k, None) for k in defaults}

    for def_k, def_v in defaults.items():
        out[def_k] = {
            k: (v if (out[def_k] is None or k not in out[def_k]) else out[def_k][k])
            for k, v in def_v.items()
        }

    return tuple([v for v in out.values()])



"""
Testing suite for the :py:mod:`pyXMIP.databases` module.
"""
import os

import pytest
from astropy import units
from astropy.coordinates import SkyCoord

from pyXMIP.structures.databases import DEFAULT_DATABASE_REGISTRY, RemoteDatabase
from pyXMIP.tests.utils import check_astropy_table

database_answer_subdir = "database_answers"
query_radius_check_position = SkyCoord(ra=1, dec=3, unit="deg")
radius = 2 * units.arcmin


@pytest.mark.parametrize(
    "database",
    [DEFAULT_DATABASE_REGISTRY[j] for j in DEFAULT_DATABASE_REGISTRY.remotes],
)
def test_radius_query(database: RemoteDatabase, answer_store, answer_dir):
    """
    Queries each database for a specified radius and position. Checks against saved answer tables.
    """
    query = database.query_radius(query_radius_check_position, radius=radius)
    check_astropy_table(
        query,
        answer_dir,
        os.path.join(database_answer_subdir, f"{database.name}_query_table.fits"),
        answer_store,
    )


class RemoteDatabaseTest:
    """
    General class for setting up and accessing testing resources for the database tests.
    """

    remote_database_name = None  # The name of the remote database.



"""
Pytest configuration for the ``pyXMIP`` testing suite.

Notes
-----

There are a few command line arguments that can be provided when running the testing suite:

- ``answer_dir`` (Required) The directory in which the answers are stored.
- ``answer_store`` If this flag is active, we create new answers in ``answer_dir`` and don't check answers.
- ``runslow`` If active, slow tests will also be run.
"""
import os

import pytest


def pytest_addoption(parser):
    parser.addoption("--answer_dir", help="Directory where answers are stored.")
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )
    parser.addoption(
        "--answer_store",
        action="store_true",
        help="Generate new answers, but don't test.",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    # ------------------------------------------ #
    # Managing slow skipping                     #
    # ------------------------------------------ #
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture()
def answer_store(request):
    """fetches the ``--answer_store`` option."""
    return request.config.getoption("--answer_store")


@pytest.fixture()
def answer_dir(request):
    """fetches the ``--answer_dir`` option."""
    ad = os.path.abspath(request.config.getoption("--answer_dir"))
    if not os.path.exists(ad):
        os.makedirs(ad)
    return ad



"""
Reduction processes and related structures.

Notes
-----

For guidance on working with reduction processes, we recommend looking at the guide: :ref:`cross_Referencing`.

"""
from abc import ABC, abstractmethod
from typing import Callable, Self

import numpy as np
import pandas as pd
from pydantic import BaseModel, model_validator

from pyXMIP.cross_reference import PydanticCMD
from pyXMIP.utilities.logging import mainlog
from pyXMIP.utilities.sql import chunk_sql_query_operation
from pyXMIP.utilities.types import ICRSCoordinateStdErrorSpecifier, TableColumn


class ReductionProcess(BaseModel, ABC):
    """
    Abstract class structure for reduction processes.
    """

    # ------------------------------ #
    # Schema Specifiers              #
    # ------------------------------ #
    settings_header: str = None
    """str: The settings header under which to find the parameters for this reduction process.

    When initializing a reduction process or set of reduction processes from a schema, the schema's underlying data dictionary
    will contain all of the relevant parameters under a header (i.e. ``ABSTRACT_PROCESS_PARAMS``). This attribute allows ``pyXMIP``
    to identify the :py:class:`ReductionProcess` associated with the settings and set everything up correctly.
    """
    settings_flag: str = None
    """str: The schema flag in ``RUNTIME_PARAMS`` which marks this process as enabled.
    """
    # ------------------------------ #
    # Process Specifiers             #
    # ------------------------------ #
    process_name: str
    """str: The name of this reduction process.
    """
    cross_match_database: PydanticCMD
    """:py:class:`cross_reference.CrossMatchDatabase`: The CMD this reduction process will run on.
    """
    table: str
    """str: The table name this reduction process will run on."""
    fill_unknown: bool = False
    """bool: If ``True``, then process will attempt to generate values for the process."""

    def __init__(self, **kwargs):
        """
        Initialize the reduction process.

        Parameters
        ----------
        process_name: str
            The name of the final :py:class:`ReductionProcess` instance.
        **kwargs:
            parameter and assigned value pairs corresponding to the parameters of this reduction process.

        Notes
        -----
        The initialization process simply assigns the ``process_name`` to ``self.name`` and adds the kwargs to the
        ``self._parameter_dictionary``, which is then accessed by the descriptors. Parameters are validated during instantiation.

        """
        super().__init__(**kwargs)

    def __str__(self):
        return f"<ReductionProcess: {self.process_name}>"

    def __repr__(self):
        return self.__str__()

    @property
    def debug_flag(self) -> str:
        """
        The string annotation used in debugging statements for this process.
        """
        return f"[{self.process_name}|{self.table}]"

    @property
    def proc_tmp_table(self) -> str:
        """str: The temp table name where process results are tabulated."""
        return f"{self.process_name}_TMP"

    @property
    def op_tmp_table(self) -> str:
        """str: The temp table where the newly joined outputs are stored while replace occurs."""
        return f"{self.table}_TMP"

    @property
    def obj_col(self) -> str:
        """str: The column of the ``table`` containing the object identifier."""
        return self.cross_match_database.schema.table_schema[self.table].NAME

    @property
    def score_col(self) -> str:
        """str: The column name storing the scores for the process."""
        return f"{self.process_name}_SCORE"

    @model_validator(mode="after")
    def validate(self) -> Self:
        """
        Custom validator for this reduction process.

        The ``validate`` method is a standard ``pydantic`` validator (runs after ``__init__`` and returns ``self``) which
        parses the provided attributes and utilizes the :py:attr:`ReductionProcess.cross_match_database` and :py:attr:`ReductionProcess.table`
        values to fill in other missing values. This method should only succeed if there is enough information to generate the operator and
        SQL query.
        """
        # ----------------------------------------------- #
        # Constructing Missing Settings                   #
        # ----------------------------------------------- #
        if self.fill_unknown:
            # The unknowns are going to be filled.
            mainlog.info(
                f"{self.debug_flag} Attempting to generate any missing attributes."
            )
            self._fill_unknown_parameters()

        # ----------------------------------------------- #
        # Validating the Model                            #
        # ----------------------------------------------- #
        mainlog.info(f"{self.debug_flag} Validating model parameters.")

        # -- default checks -- #
        assert (
            self.table in self.cross_match_database.match_tables
        ), f"{self.debug_flag} TABLE {self.table} not in {self.cross_match_database}."

        # Non-standard checks
        self._validate()

        return self

    @abstractmethod
    def _fill_unknown_parameters(self):
        pass

    @abstractmethod
    def _validate(self):
        return self

    @abstractmethod
    def generate_sql_query(self) -> str:
        """
        Generate the correct ``sql`` query for this reduction process.

        Returns
        -------
        str
            The output SQL query.
        """
        pass

    @abstractmethod
    def generate_operator(self) -> Callable[[pd.DataFrame], pd.DataFrame | pd.Series]:
        """
        Generate the correct ``operator`` function for this reduction process.

        Returns
        -------
        Callable[[pd.DataFrame],pd.DataFrame|pd.Series]
            The function which operates on the table.

        Notes
        -----
        The output operator is a function with signature ``func(table: pd.DataFrame) -> pd.DataFrame``. The output dataframe
        **must** have 3 columns (``"<REDUCTION_PROCESS_NAME>_SCORE"``) with ``float`` values corresponding to the respective scores
        of each entry in the input table as well as the ``CATOBJ`` column specifying the catalog object and the column identifying
        the potential match.
        """
        pass

    def __call__(self, *args, overwrite=False, **kwargs):
        import sqlalchemy as sql

        # ------------------------------------- #
        # Setup and validation                  #
        # ------------------------------------- #
        tqdm_kwargs = {"desc": self.debug_flag}
        _sql_engine = self.cross_match_database._sql_engine

        # -- check that we can actually perform this reduction -- #
        _check_status = self.cross_match_database.check_meta(
            self.process_name, self.table
        )

        if _check_status and overwrite:
            # The process has already been run, but overwrite = True.
            with _sql_engine.connect() as conn:
                mainlog.warning(
                    f"{self.debug_flag} Process has already been performed. Overwritting."
                )
                conn.execute(
                    sql.text(
                        f"ALTER TABLE `{self.table}` DROP COLUMN `{self.process_name}_SCORE`"
                    )
                )
        elif _check_status and not overwrite:
            raise ValueError(
                f"{self.debug_flag} The process has already been performed and overwrite=False."
            )
        else:
            pass

        # -- construct the query and the operator -- #
        # These call to the two abstract generator methods for the setup procedure.
        query, _operator = self.generate_sql_query(), self.generate_operator()

        # ------------------------------------- #
        # Create the callable operator          #
        # ------------------------------------- #
        # The chunked SQL operator is called to produce a version of OPERATOR which runs efficiently on
        # chunks.
        #
        # The new signature will be f(engine,query,output_table,*args,**kwargs)
        operator = chunk_sql_query_operation(**dict(tqdm_kwargs=tqdm_kwargs))(_operator)

        # -------------------------------------- #
        # Running the operator                   #
        # -------------------------------------- #
        operator(_sql_engine, query, f"{self.process_name}_TMP", *args, **kwargs)

        # -------------------------------------- #
        # Post-Processing                        #
        # -------------------------------------- #
        self._post_process()

    def _post_process(self):
        # ---------------------------------------------- #
        # Constructing queries                           #
        # ---------------------------------------------- #
        import sqlalchemy as sql

        JOIN_QUERY = sql.text(
            f"CREATE TABLE {self.op_tmp_table} AS "
            f"SELECT {self.table}.* , {self.proc_tmp_table}.`{self.score_col}` "
            f"FROM {self.table} LEFT JOIN {self.proc_tmp_table} "
            f"ON (({self.table}.CATOBJ = {self.proc_tmp_table}.CATOBJ) AND "
            f"({self.table}.`{self.obj_col}` = {self.proc_tmp_table}.`{self.obj_col}`))"
        )

        DROP_QUERY = sql.text(f"DROP TABLE {self.table}")
        RENAME_QUERY = sql.text(
            f"ALTER TABLE {self.op_tmp_table} RENAME TO {self.table}"
        )

        with self.cross_match_database._sql_engine.connect() as conn:
            mainlog.info(
                f"{self.debug_flag} Cleaning up SQL components... (Joining tables)"
            )
            conn.execute(JOIN_QUERY)
            mainlog.info(
                f"{self.debug_flag} Cleaning up SQL components... (Dropping tables)"
            )
            conn.execute(DROP_QUERY)
            mainlog.info(
                f"{self.debug_flag} Cleaning up SQL components... (Renaming tables)"
            )
            conn.execute(RENAME_QUERY)

        # -- Managing the META data -- #
        self.cross_match_database.meta_add(
            self.process_name, self.table, is_reduction=True
        )


class AstrometricReductionProcess(ReductionProcess):
    """
    Reduction process for refining match confidence on the basis of astrometric precision and separation
    between sources.
    """

    settings_header: str = "ASTROMETRY_PARAMS"
    settings_flag: str = "enable_astrometry"
    process_name: str = "ASTROMETRIC_REDUCTION"

    # ----------------------------------------------------------- #
    # Settings                                                    #
    # ----------------------------------------------------------- #
    CATALOG_ERR: ICRSCoordinateStdErrorSpecifier | None = None
    """:py:class:`utilities.types.CoordinateStdErrorSpecifier`: The error specifier for the catalog.
    """
    DATABASE_ERR: ICRSCoordinateStdErrorSpecifier | None = None
    """:py:class:`utilities.types.CoordinateStdErrorSpecifier`: The error specifier for the database.
    """
    prior: Callable[[], float] = lambda x: 1 / np.array(x["NMATCH"])

    def _get_displacement_op(
        self,
    ) -> Callable[[pd.DataFrame, ...], tuple[np.ndarray, np.ndarray]]:
        """
        Constructs the displacement operator for this reduction process
        """
        from astropy.coordinates import SkyCoord
        from astropy.units import Quantity

        # ---------------------------------------------------- #
        # Construct the names for important columns            #
        # ---------------------------------------------------- #
        # Necessary columns are the positions and units for each of the
        # coordinate positions.
        _cra, _cdec, _dra, _ddec = "CATRA", "CATDEC", "DBRA", "DBDEC"
        _cra_unit, _cdec_unit = "deg", "deg"
        _dra_unit, _ddec_unit = (
            self.cross_match_database.schema.table_schema[
                self.table
            ].column_map.RA.unit,
            self.cross_match_database.schema.table_schema[
                self.table
            ].column_map.DEC.unit,
        )

        # ---------------------------------------------------- #
        # Construct the operator                               #
        # ---------------------------------------------------- #
        def _coord_func(
            table_chunk: pd.DataFrame,
            cra=_cra,
            cdec=_cdec,
            dra=_dra,
            ddec=_ddec,
            cra_unit=_cra_unit,
            cdec_unit=_cdec_unit,
            dra_unit=_dra_unit,
            ddec_unit=_ddec_unit,
        ) -> tuple[np.ndarray, np.ndarray]:
            # -- generate the coordinate positions -- #
            cp = SkyCoord(
                ra=Quantity(table_chunk[cra], unit=cra_unit),
                dec=Quantity(table_chunk[cdec], unit=cdec_unit),
            )
            dp = SkyCoord(
                ra=Quantity(table_chunk[dra], unit=dra_unit),
                dec=Quantity(table_chunk[ddec], unit=ddec_unit),
            )

            # -- return the separations -- #
            return (cp.ra - dp.ra).to_value("arcsec"), (cp.dec - dp.dec).to_value(
                "arcsec"
            )

        return _coord_func

    def generate_operator(self) -> Callable[[pd.DataFrame], pd.DataFrame | pd.Series]:
        # ----------------------------------- #
        # Get the column separation operator  #
        # ----------------------------------- #
        # the displacement_function maps the table to the separation between RA and DEC coords.
        displacement_function = self._get_displacement_op()

        # ------------------------------------ #
        # Construct main operator              #
        # ------------------------------------ #
        # There are 3 separate formulae depending on the nature of the specified errors.
        if self.DATABASE_ERR is None or (self.DATABASE_ERR.mode is None):
            # The database is assumed to be ideal. Thus, we model error domain as a single gaussian.
            crau, cdecu = (
                self.CATALOG_ERR.lon_error.unit,
                self.CATALOG_ERR.lat_error.unit,
            )

            # -- Pull additional meta-data -- #
            def _func(
                table_chunk: pd.DataFrame,
                df: Callable[
                    [pd.DataFrame, ...], tuple[np.ndarray, np.ndarray]
                ] = displacement_function,
            ):
                # -- compute displacement vector -- #
                ra_dis, dec_dis = df(table_chunk)
                # -- compute error values -- #
                sra, sdec = (
                    (np.array(table_chunk["CATRA_ERR"]) * crau).to_value("arcsec"),
                    (np.array(table_chunk["CATDEC_ERR"]) * cdecu).to_value("arcsec"),
                )

                coef = ((2 * np.pi) ** (-1)) * (sra * sdec) ** (-1)
                table_chunk["bf"] = coef * np.exp(
                    (-0.5) * ((ra_dis / sra) ** 2 + (dec_dis / sdec) ** 2)
                )

                return table_chunk

        elif self.CATALOG_ERR is None or (self.CATALOG_ERR.mode is None):
            # The catalog is assumed to be ideal. Thus, we model error domain as a single gaussian.
            crau, cdecu = (
                self.DATABASE_ERR.lon_error.unit,
                self.DATABASE_ERR.lat_error.unit,
            )

            # -- Pull additional meta-data -- #
            def _func(table_chunk, df=displacement_function):
                # -- compute displacement vector -- #
                ra_dis, dec_dis = df(table_chunk)
                # -- compute error values -- #
                sra, sdec = (
                    (np.array(table_chunk["DBRA_ERR"]) * crau).to_value("arcsec"),
                    (np.array(table_chunk["DBDEC_ERR"]) * cdecu).to_value("arcsec"),
                )

                coef = ((2 * np.pi) ** (-1)) * (sra * sdec) ** (-1)
                table_chunk["bf"] = coef * np.exp(
                    (-0.5) * ((ra_dis / sra) ** 2 + (dec_dis / sdec) ** 2)
                )
                return table_chunk

        else:
            # -- Complex formalism -- #
            crau, cdecu = (
                self.CATALOG_ERR.lon_error.unit,
                self.CATALOG_ERR.lat_error.unit,
            )
            drau, ddecu = (
                self.DATABASE_ERR.lon_error.unit,
                self.DATABASE_ERR.lat_error.unit,
            )

            def _func(table_chunk, df=displacement_function):
                # -- compute displacement vector -- #
                ra_dis, dec_dis = df(table_chunk)

                # -- compute error values -- #
                cra, cdec, dra, ddec = (
                    (np.array(table_chunk["CATRA_ERR"]) * crau).to_value("arcsec"),
                    (np.array(table_chunk["CATDEC_ERR"]) * cdecu).to_value("arcsec"),
                    (np.array(table_chunk["DBRA_ERR"]) * drau).to_value("arcsec"),
                    (np.array(table_chunk["DBDEC_ERR"]) * ddecu).to_value("arcsec"),
                )
                iscra, iscdec, isdra, isddec = (
                    cra**-2,
                    cdec**-2,
                    dra**-2,
                    ddec**-2,
                )

                coef = 1 / (
                    (2 * np.pi)
                    * (
                        (cra * cdec * dra * ddec)
                        / (np.sqrt((cra**-2 + dra**-2) * (cdec**-2 + ddec**-2)))
                    )
                )

                exp = (-0.5) * (
                    ((iscra * isdra) / (iscra + isdra)) * ra_dis**2
                    + ((iscdec * isddec) / (iscdec + isddec)) * dec_dis**2
                )

                table_chunk["bf"] = coef * np.exp(exp)
                return table_chunk

        # ------------------------------------------------ #
        # Implement the prior                              #
        # ------------------------------------------------ #
        def _pfunc(table_chunk, f=_func, df=displacement_function, p=self.prior):
            table_chunk = f(table_chunk, df=df)

            table_chunk[self.score_col] = (
                1 + (1 - p(table_chunk)) / (p(table_chunk) * table_chunk["bf"])
            ) ** (-1)

            return table_chunk

        return _pfunc

    def generate_sql_query(self) -> str:
        # ------------------------------------------------ #
        # Adding fixed columns to the column list          #
        # ------------------------------------------------ #
        # This includes the object identifier columns for both db and CATALOG along with the
        # coordinate columns in ICRS.
        columns = {
            "CATOBJ": "CATALOG.`CATOBJ`",
            f"{self.obj_col}": f"{self.table}.`{self.obj_col}`",
            "CATRA": "CATALOG.`RA`",
            "CATDEC": "CATALOG.`DEC`",
            "DBRA": f"{self.table}.`RA`",  # ALL tables have RA; its a fixed column.
            "DBDEC": f"{self.table}.`DEC`",  # ALL tables have DEC; its a fixed column.
            "NMATCH": f"{self.table}.`CATNMATCH`",
        }
        # ------------------------------------------------ #
        # Adding un-fixed columns to the column list       #
        # ------------------------------------------------ #
        for tbl, error_struct in zip(
            [self.table, "CATALOG"], [self.DATABASE_ERR, self.CATALOG_ERR]
        ):
            # Add coordinate error columns where necessary.
            if error_struct is None or error_struct.mode is None:
                # The error structure is None, we can just skip.
                continue
            elif error_struct.mode == "circular":
                # The error structure is circular. We need to check if the error specifier is a value or a columns.
                if isinstance(error_struct.position_error, TableColumn):
                    # this is a column and needs to be added.
                    columns[
                        "DBRA_ERR" if tbl != "CATALOG" else "CATRA_ERR"
                    ] = f"{tbl}.`{error_struct.position_error.name}`"
                    columns[
                        "DBDEC_ERR" if tbl != "CATALOG" else "CATDEC_ERR"
                    ] = f"{tbl}.`{error_struct.position_error.name}`"
                else:
                    pass
            else:
                # the error structure is axial.
                if isinstance(error_struct.lat_error, TableColumn):
                    columns[
                        "DBDEC_ERR" if tbl != "CATALOG" else "CATDEC_ERR"
                    ] = f"{tbl}.`{error_struct.lat_error.name}`"
                if isinstance(error_struct.lon_error, TableColumn):
                    columns[
                        "DBRA_ERR" if tbl != "CATALOG" else "CATRA_ERR"
                    ] = f"{tbl}.`{error_struct.lon_error.name}`"

        # -- Now construct the SQL query. -- #
        column_string = ", ".join([f"{v} as `{k}`" for k, v in columns.items()])

        SQL_QUERY = (
            "SELECT %(columns)s FROM %(table)s LEFT JOIN CATALOG USING(CATOBJ)"
            % dict(columns=column_string, table=self.table)
        )
        return SQL_QUERY

    def _validate(self):
        # -- Check that at least one of the processes is non-redundant -- #
        if (self.DATABASE_ERR is None) and (self.CATALOG_ERR is None):
            raise ValueError(
                f"{self.debug_flag} Neither component of the match has specified error."
            )

        for err_spec in [self.DATABASE_ERR, self.CATALOG_ERR]:
            if err_spec is None:
                continue

            err_spec.check_empty()

    def _fill_unknown_parameters(self):
        error_specifiers = [self.DATABASE_ERR, self.CATALOG_ERR]
        schemas = [
            self.cross_match_database.schema.table_schema[self.table],
            self.cross_match_database.schema.catalog_schema,
        ]

        for error_specifier, schema in zip(error_specifiers, schemas):
            # -- Skip unspecified error specifiers -- #
            # This way, the user is always trusted to ignore a component if they want to.
            if error_specifier is None:
                continue

            # -- Seek error columns in the schema -- #
            if schema.column_map.RA_ERR is not None:
                mainlog.info(
                    f"{self.debug_flag} Filled parameter lon_err with {schema.column_map.RA_ERR.name}"
                )
                error_specifier.lon_error = schema.column_map.RA_ERR
            if schema.column_map.DEC_ERR is not None:
                mainlog.info(
                    f"{self.debug_flag} Filled parameter lat_err with {schema.column_map.DEC_ERR.name}"
                )
                error_specifier.lat_error = schema.column_map.DEC_ERR


if __name__ == "__main__":
    # load the cross match database from disk
    from pyXMIP.cross_reference import CrossMatchDatabase

    cmd = CrossMatchDatabase("../../docs/source/examples/data/cross_matched.db")
    cmd.schema.catalog_schema.column_map.RA_ERR = {"name": "RA_UPERR", "unit": "arcsec"}
    cmd.schema.catalog_schema.column_map.DEC_ERR = {
        "name": "DEC_UPERR",
        "unit": "arcsec",
    }
    proc = AstrometricReductionProcess(
        cross_match_database=cmd,
        table="NED_STD_MATCH",
        CATALOG_ERR={},
        fill_unknown=True,
    )
    print(proc.CATALOG_ERR)



"""
SQL interaction module for ``pyXMIP``.
"""
from typing import Callable, Collection

import pandas as pd
import sqlalchemy as sql
from tqdm.asyncio import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from pyXMIP.utilities.logging import mainlog


def _count_table(engine: sql.Engine, table: str) -> int:
    with engine.connect() as conn:
        return int(conn.execute(sql.text(f"SELECT COUNT(*) FROM {table}")).scalar())


def chunk_sql_query_operation(tqdm_kwargs: dict = None):
    """
    Meta-decorator function to perform operations on SQL queries in chunks instead of loading the entire database into
    memory at once.

    Parameters
    ----------
    tqdm_kwargs: dict
        The arguments to pass to tqdm during runtime.

    Returns
    -------
    Callable
        The resulting decorator.

    """

    if tqdm_kwargs is None:
        tqdm_kwargs = {}

    def _chunk_sql_query_operation(
        function: Callable[[pd.DataFrame, ...], pd.DataFrame]
    ):
        """
        Decorator function to perform operations on SQL queries in chunks instead of loading the entire database into
        memory at once.

        Parameters
        ----------
        function: Callable
            The operation to perform on each of the query chunks.

            The signature of ``function`` should be ``func(table: pd.DataFrame, *args, **kwargs)``. ``*args`` and ``**kwargs`` are
            then provided at runtime by the user. (See the notes on the returned function)

        Returns
        -------
        Callable
            The resulting decorator output.

            This is a function with signature ``func(engine: sql.Engine, sql_query: str, otable_name: str,chunksize:int = 1000, *args, **kwargs)``.
            The ``engine`` must be the SQL engine to connect to, the ``sql_query`` is the query to execute and the result is then operated on
            by the wrapped function. Finally, ``otable_name`` specifies what name to provide to the output table. ``*args,**kwargs`` are passed
            directly to ``function``.

        """

        def base_wrapper(
            engine: sql.Engine,
            sql_query: str,
            otable_name: str,
            *args,
            **kwargs,
        ) -> None:
            """
            Perform a chunkwise operation on a particular ``sql_query`` and write the resulting output to disk.

            Parameters
            ----------
            engine: sql.Engine
                The ``sqlalchemy`` engine connection to the underlying database.
            sql_query: str
                The sql query to run.
            otable_name: str
                The name to give to the output table.
            chunksize: int, optional
                The maximum allowed chunksize for each operation.
            args:
                Additional arguments to pass to the underlying function.
            kwargs:
                Additional keyword arguments to pass to the underlying function.
            """
            chunksize = kwargs.pop("chunksize", 1000)

            with engine.connect() as conn, logging_redirect_tqdm(loggers=[mainlog]):
                # Connect to the SQL engine and redirect logging via tqdm.

                for chunk_id, sql_chunk in enumerate(
                    tqdm(
                        pd.read_sql(sql_query, conn, chunksize=chunksize),
                        **tqdm_kwargs,
                    )
                ):
                    # Read the SQL query in chunks of specified chunksize. Operate on those chunks with the specified function.
                    # -- perform the function -- #
                    chunk_image = function(*(sql_chunk, *args), **kwargs)

                    # -- Writing the output to the SQL database -- #
                    # If chunk_id == 0, then we need to create the table,
                    # otherwise, we just append to the existing table.

                    if chunk_id == 0:
                        # -- This database table doesn't exist yet. -- #

                        chunk_image.to_sql(
                            otable_name, conn, if_exists="replace", index=False
                        )
                    else:
                        # -- Write -- #
                        chunk_image.to_sql(
                            otable_name, conn, if_exists="append", index=False
                        )

                    conn.commit()

        return base_wrapper

    return _chunk_sql_query_operation


def chunk_sql_table_operation(
    columns: Collection[str] | str = "*",
    inplace: bool = False,
    tqdm_kwargs: dict = None,
):
    """
    Meta-decorator function to perform operations on SQL tables in chunks instead of loading the entire database into
    memory at once.

    Parameters
    ----------
    tqdm_kwargs: dict
        The arguments to pass to tqdm during runtime.
    columns: str or Collection[str], optional
        The columns to include in the query. By default, this is ``"*"`` which will cause the function to operate
        on the complete table with all columns.
    inplace: bool, optional
        If ``True``, then (instead of creating a table ``"{table}_PROCESSED"``, the original table (``table``) will be replaced
        with the output of this operation.

    Returns
    -------
    Callable
        The resulting decorator.

    """
    if tqdm_kwargs is None:
        tqdm_kwargs = {}

    def _chunk_sql_table_operation(
        function: Callable[[pd.DataFrame, ...], pd.DataFrame]
    ):
        """
        Decorator function to perform operations on SQL table in chunks instead of loading the entire database into
        memory at once.

        Parameters
        ----------
        function: Callable
            The operation to perform on each of the query chunks.

            The signature of ``function`` should be ``func(table: pd.DataFrame, *args, **kwargs)``. ``*args`` and ``**kwargs`` are
            then provided at runtime by the user. (See the notes on the returned function)

        Returns
        -------
        Callable
            The resulting decorator output.

            This is a function with signature ``func(engine: sql.Engine, table: str,chunksize:int = 1000, *args, **kwargs)``.
            The ``engine`` must be the SQL engine to connect to. ``*args,**kwargs`` are passed directly to ``function``.

        """
        # The base_wrapper needs to be converted into a new wrapper with the desired (simplified) signature.
        # This means that the sql query must be constructed and the otable generated.

        def wrapper(engine: sql.Engine, table: str, *args, **kwargs):
            """
            Perform a chunkwise operation on a particular ``sql_query`` and write the resulting output to disk.

            Parameters
            ----------
            engine: sql.Engine
                The ``sqlalchemy`` engine connection to the underlying database.
            table: str
                The sql table to run on.
            chunksize: int, optional
                The maximum size of the operation chunks.
            args:
                Additional arguments to pass to the underlying function.
            kwargs:
                Additional keyword arguments to pass to the underlying function.
            """
            # -- Construct the basic wrapper from the general decorator -- #

            N = (_count_table(engine, table) // kwargs.get("chunksize", 1000)) + 1
            tqdm_kwargs["total"] = N

            # Now we can create the wrapper.
            base_wrapper = chunk_sql_query_operation(tqdm_kwargs=tqdm_kwargs)(function)

            # -- Constructing the SQL query for the specified table -- #
            sql_query = "SELECT %(columns)s FROM %(table)s" % dict(
                columns=(columns if isinstance(columns, str) else ", ".join(columns)),
                table=table,
            )
            otable_name = (
                f"{table}_PROCESSED"  # Standard prescription for the output table.
            )

            # -- Execute the base_wrapper -- #
            base_wrapper(*(engine, sql_query, otable_name, *args), **kwargs)

            # -- Cleanup and post-processing -- #
            if inplace:
                with engine.connect() as conn:
                    # -- replace the tables -- #
                    conn.execute(sql.text(f"DROP TABLE {table}"))
                    conn.execute(
                        sql.text(f"ALTER TABLE {otable_name} RENAME TO {table}")
                    )

        return wrapper

    return _chunk_sql_table_operation



"""
Statistical Utility functions
"""
import numpy as np


def uniform_sample_spherical(n_points):
    r"""
    Return a uniform sample from a spherical surface.

    Parameters
    ----------
    n_points: int
        The number of samples to draw.

    Returns
    -------
    phi
        The azimuthal positions (in rad)
    theta
        The elevation angles (in rad)

    Notes
    -----

    This function relies on functional transform to produce the random sample. A coordinate segment of a spherical surface :math:`d\Omega` has
    area

    .. math::

        d\Omega = \sin(\theta) d\phi d\theta

    Thus, if the probability of finding :math:`k` points in a region of size :math:`d\Omega`, then

    .. math::

        P(k|d\Omega) \sim U(d\Omega) \sim \sin(\theta) U(d\phi d\theta) \sim \alpha d\psi d\omega.

    To accomplish such a transformation, we require only that

    .. math::

        \psi(\phi,\theta) = \phi,\; \omega(\phi,\theta) = \sin^{-1} \theta

    Thus, for the distribution to be independent of position we simply let

    .. math::

        \cos \phi \sim U(N).
    """
    return 2 * np.pi * np.random.uniform(0, 1, size=n_points), np.arccos(
        2 * np.random.uniform(0, 1, size=n_points) - 1
    )



from pyXMIP.cross_reference import CrossMatchDatabase, cross_match, cross_match_table
from pyXMIP.structures.databases import NED, SIMBAD, LocalDatabase, RemoteDatabase
from pyXMIP.structures.table import SourceTable, load
